#!/usr/bin/env python
#import hotshot, hotshot.stats
import os,os.path,re,logging,sys,struct,math,traceback,getopt,inspect
import types
import pygtk
pygtk.require('2.0')
import gtk,gtk.glade,gobject
import aksy

# our stuff

from ak import *
from ak.samplerobject import *
from UI import *

from utils import midiutils, modelutils

from postmod.itx import *

from aksy.device import Devices

__author__ = 'Joseph Misra'
__version__ = '0.01'


class CellRendererSamplerObject(gtk.GenericCellRenderer):
    __gproperties__ = {
            "samplerobject": (object, "samplerobject", "samplerobject", gobject.PARAM_READWRITE),
    }

    def __init__(self, s):
        self.__gobject_init__()
        gtk.GenericCellRenderer.__init__(self)
        self.s = s
        self.custom = None
        self.widget_type = None

    def do_set_property(self, pspec, value):
        setattr(self, pspec.name, value)

    def do_get_property(self, pspec):
        return getattr(self, pspec.name)

    def on_render(self, window, widget, background_area, cell_area, expose_area, flags):
        if self.custom:
            return self.custom.draw(window, widget, cell_area)
        else:
            return False

    def on_get_size(self, widget, cell_area):
        if self.custom:
            return (0,0,15,30) 
        else:
            return (0,0,0,0)

    def on_activate(event, widget, path, background_area, cell_area, flags):
        return

    def on_start_editing(event, widget, path, background_area, cell_area, flags):
        return

class CellRendererKnob(CellRendererSamplerObject):
    def __init__(self, s):
        CellRendererSamplerObject.__init__(self, s)

    def do_set_property(self, pspec, value):
        if pspec.name == "samplerobject" and not self.custom:
            self.custom = AkKnobWidget(value, "level", -600, 60, 10, "db", "AMPLITUDE")

        setattr(self, pspec.name, value)

gobject.type_register(CellRendererSamplerObject)

# create a window an add a treeview
# build model to hold program -> keygroup -> samples


class MainProgramsWindow(MainWindow):
    def __init__(self, z48):
        gtk.Window.__init__(self)
        self.set_title("ak.py")
        self.s = z48
        self.iter_dict = {}

        self.treeview = gtk.TreeView()

        self.connect("configure_event", self.on_configure_event)

        self.init()

    def build_model(self):
        self.treestore = gtk.TreeStore(str, str, object) 
        
        for program_info in self.s.programs:
            program_handle, program_name = program_info 
            p = ak.program(self.s,program_name)
            program_iter = self.treestore.append(None,[program_name,'',p])
            no_keygroups = self.s.programtools.get_no_keygroups()
            self.iter_dict[self.treestore.get_string_from_iter(program_iter)] = p
            for i in range(no_keygroups):
                kg = ak.keygroup(p,i)
                keygroup_iter = self.treestore.append(program_iter,['Keygroup ' + str(i), "--------", kg])
                self.iter_dict[self.treestore.get_string_from_iter(keygroup_iter)] = kg
                for j in range(4):
                    z = ak.zone(kg, j+1)
                    zone_iter = self.treestore.append(keygroup_iter,['Zone ' + str(j+1), z.sample, z])
                    self.iter_dict[self.treestore.get_string_from_iter(zone_iter)] = z

        return self.treestore

    def init(self):
        # based on list build TreeView
        MainProgramsWindow.do_lists(self.s)
        #self.lcd = UI.LCDScreen(self.s)

        self.xml = gtk.glade.XML("menu.glade", "mainmenu")
        self.mainmenu = self.xml.get_widget("mainmenu")
        self.xml.signal_autoconnect(self)

        self.set_size_request(500,500)

        vbox = gtk.VBox(False,0)
        vbox.pack_start(self.mainmenu)
        vbox.set_child_packing(self.mainmenu, False, True, 0, gtk.PACK_START)

        self.hpaned = gtk.HPaned()
        self.hpaned.set_position(250)
        self.scrolledwindow = gtk.ScrolledWindow()
        self.treestore = self.build_model()
        self.treeview = gtk.TreeView(self.treestore)
        self.treeview.connect("button_press_event", self.on_treeview_button_press_event)
        self.treeview.connect("button_release_event", self.on_treeview_button_release_event)
        self.treeview.connect("motion_notify_event", self.on_treeview_motion_notify_event)

        textedit = gtk.CellRendererText()
        textedit.set_property('editable', False)
        textedit.connect('edited', self.on_textedit_changed, (self.treestore, 0))

        sampleedit = gtk.CellRendererCombo()
        sampleedit.set_property('editable', True)
        sampleedit.set_property('model', self.s.samplesmodel)
        sampleedit.set_property('text-column', 1)
        sampleedit.connect('edited', self.on_sampleedit_changed, (self.treestore, 1))
        sampleedit.connect('editing-started', self.on_sampleedit_started)

        leveledit = CellRendererKnob(self.s)
        """
        tuneedit = gtk.GenericCellRenderer()
        panedit = gtk.GenericCellRenderer()
        filteredit = gtk.GenericCellRenderer()
        """

        self.treeview.append_column(gtk.TreeViewColumn("Name", textedit, text=0))
        self.treeview.append_column(gtk.TreeViewColumn("Sample", sampleedit, text=1))
        self.treeview.append_column(gtk.TreeViewColumn("Level", leveledit, samplerobject=2))
        """
        self.treeview.append_column(gtk.TreeViewColumn("Tune", tuneedit))
        self.treeview.append_column(gtk.TreeViewColumn("Pan", panedit))
        self.treeview.append_column(gtk.TreeViewColumn("Filter", filteredit))
        """

        self.scrolledwindow.add(self.treeview)
        self.hpaned.add1(self.scrolledwindow)
        self.hpaned.add2(gtk.DrawingArea())
        vbox.pack_start(self.hpaned)
        self.add(vbox)

        self.queue_draw()
        self.show_all()

    def call_cell_func(self, widget, event, which):
        result = self.treeview.get_path_at_pos(int(event.x), int(event.y))

        if result:
            path, column, x, y = result

            renderers = column.get_cell_renderers() 

            area = self.treeview.get_cell_area(path, column)

            if area.height > 0:
                #print path, event.x, event.y, area.x, area.y, area.width, area.height

                e = event.copy()
                e.x = event.x - float(area.x)
                e.y = event.y - float(area.y)

                for renderer in renderers:
                    if type(renderer) is CellRendererKnob:
                        if which == "on_button_press":
                            renderer.custom.on_button_press(widget, e)
                        elif which == "on_motion_notify":
                            renderer.custom.on_motion_notify_event(widget, e)
                        elif which == "on_button_release":
                            renderer.custom.on_button_release(widget, e)

        self.treeview.queue_draw()
        return False

    def on_treeview_motion_notify_event(self, widget, event):
        self.call_cell_func(widget, event, "on_motion_notify")
        return False

    def on_treeview_button_press_event(self, widget, event):
        self.call_cell_func(widget, event, "on_button_press")
        return False

    def on_treeview_button_release_event(self, widget, event):
        self.call_cell_func(widget, event, "on_button_release")
        return False

    def on_sampleedit_started(self, cell, editable, path):
        if not self.treestore[path][0].startswith('Zone'):
            cell.stop_editing(True)
            return True

    def on_sampleedit_changed(self, cell, path, new_text, user_data):
        model, index = user_data
        model[path][index] = new_text
        if path in self.iter_dict:
            so = self.iter_dict[path]
            if type(so) is ak.zone:
                so.set("sample", new_text)
        else:
            print "cant find", path, "in dict"
        return 

    def on_textedit_changed(self, cell, path, new_text, user_data):
        model, index = user_data
        model[path][index] = new_text
        return 

    @staticmethod
    def do_list_helper(tools):
        samples = []
        samples_tmp = tools.get_handles_names()

        i = 0

        while i < (len(samples_tmp)):
            samples.append([samples_tmp[i],samples_tmp[(i+1)]]) 
            i += 2 

        return samples

    @staticmethod
    def do_lists(s):
        samples = MainProgramsWindow.do_list_helper(s.sampletools)
        programs = MainProgramsWindow.do_list_helper(s.programtools)
        multis = MainProgramsWindow.do_list_helper(s.multitools)

        setattr(s,'samples',samples)
        setattr(s,'programs',programs)
        setattr(s,'multis',multis)

        if len(s.samples) == 0:
            print "No samples..."

        if len(s.programs) == 0:
            print "No programs..."

        if len(s.multis) == 0:
            print "No multis..."

        setattr(s,'samplesmodel', modelutils.get_model_from_list(s.samples))
        setattr(s,'programsmodel', modelutils.get_model_from_list(s.programs))
        setattr(s,'multismodel', modelutils.get_model_from_list(s.multis))

    def on_configure_event(self, widget, event):
        return False


def main(): 
    z48 = Devices.get_instance("z48", "usb")
    #z48 = None
    win = MainProgramsWindow(z48)
    win.connect("delete-event", gtk.main_quit)
    gtk.main()

if __name__ == "__main__":
    prof = hotshot.Profile("ak.py.prof")
    prof.runcall(main)
    stats = hotshot.stats.load("ak.py.prof")
    stats.sort_stats('time', 'calls')
    stats.print_stats(10)





#!/usr/bin/python

import fuse

from time import time

import stat
import os
import os.path
import errno
import logging

from aksy.device import Devices
from aksy.concurrent import transaction

from aksy.devices.akai import sampler as samplermod
from aksy import config
from aksyfs import common

fuse.fuse_python_api = (0, 2)

fuse.feature_assert('stateful_files')

MAX_FILE_SIZE_SAMPLE = 512 * 1024 * 1024 # 512 MB
MAX_FILE_SIZE_OTHER = 16 * 1024 # 16K
START_TIME = time()

LOG = logging.getLogger('aksy.aksyfs.aksyfuse')

class AksyFile(fuse.FuseFileInfo):
    @staticmethod
    def set_sampler(sampler):
        AksyFile.sampler = sampler
    
    @transaction()
    def __init__(self, path, flags, *mode):
        self.direct_io = True
        self.upload = bool(flags & os.O_WRONLY)
        self.name = os.path.basename(path)
        location = common._splitpath(path)[0]
        if location == "memory":
            self.location = samplermod.AkaiSampler.MEMORY
        elif location == "disks":
            self.location = samplermod.AkaiSampler.DISK
        else:
            raiseException(errno.ENOENT)

        self.path = common._create_cache_path(path)
        if not self.is_upload() and not common._cache_path_exists(path):
            AksyFile.sampler.transfertools.get(self.name, self.path, self.location)
        
        self.handle = os.open(self.path, flags, *mode)

    @transaction()
    def release(self, flags):
        if self.is_upload():
            try:
                AksyFile.sampler.transfertools.put(self.get_path(), None, self.get_location())
            except IOError, exc:
                # TODO: move to a method where we can raise exceptions
                LOG.exception( "Exception occurred: ", exc)
        os.close(self.handle)

    def write(self, buf, offset):
        os.lseek(self.handle, offset, 0)
        return os.write(self.handle, buf)

    def read(self, length, offset):
        os.lseek(self.handle, offset, 0)
        return os.read(self.handle, length)

    def get_name(self):
        return self.name
    
    def get_path(self):
        return self.path
    
    def is_upload(self):
        return self.upload

    def get_handle(self):
        return self.handle
    
    def get_location(self):
        return self.location

class FSStatInfo(fuse.StatVfs):
    def __init__(self, mem_total, mem_free):
        fuse.StatVfs.__init__(self)
        self.f_bsize = 1024
        self.f_frsize = 1024
        self.f_blocks = mem_total
        self.f_bfree = mem_free
        self.f_bavail = mem_free

class AksyFS(common.AksyFS, fuse.Fuse): #IGNORE:R0904
    def __init__(self, *args, **kw):
        # todo postpone sampler init
        common.AksyFS().__init__()
        
        fuse.Fuse.__init__(self, *args, **kw)
        self.multithreaded = True
        cfg = config.get_config()
        self.sampler_type = cfg.get('sampler', 'type')
        self.connector = config.get_value(cfg, 'sampler', 'connector')
        self.file_class = AksyFile
        stat_home = os.stat(os.path.expanduser('~'))
        common.StatInfo.set_owner(stat_home[stat.ST_UID], stat_home[stat.ST_GID])

    def access(self, path, mode):
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug( "**access (" + path + ") (no-op)")
    
    @transaction()
    def readdir(self, path, offset):
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug( '*** readdir(%s, %s)', path, offset)
        folder = self.cache[path]
        for child in folder.get_children():
            yield fuse.Direntry(child.get_name())

    @transaction()
    def mknod(self, path, mode, dev):
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug( 'mknod(%s, %s, %s)', path, mode, dev)
        self.cache[path] = common.FileStatInfo(path, None)
        self.get_parent(path).refresh()

    def truncate(self, path, size):
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug( "truncate (%s, %s) (no-op)", path, size)
    
    @transaction()
    def statfs(self):
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug( "statfs (metrics on memory contents only)")
        mem_total = self.sampler.systemtools.get_wave_mem_size()
        mem_free = self.sampler.systemtools.get_free_wave_mem_size()
        return FSStatInfo(mem_total, mem_free)

    def init_sampler(self, sampler):
        self.root = common.FSRoot(sampler)
        self.sampler = sampler
        self.cache = { '/': self.root }
        AksyFile.set_sampler(sampler)

    def main(self, sampler, *args, **kw):
        self.init_sampler(sampler)
        return fuse.Fuse.main(self, *args, **kw)
    
    def destroy(self):
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug('fsdestroy')
        self.sampler.close()
        

def raiseUnsupportedOperationException():
    raiseException(errno.ENOSYS)

def raiseException(err):
    raise OSError(err, 'Exception occurred')

def main():
    usage = """
Aksyfs: mount your sampler as a filesystem

""" + fuse.Fuse.fusage

    fs = AksyFS(version="%prog " + fuse.__version__,
                 usage=usage,
                 dash_s_do='setsingle')

    fs.parser.add_option(mountopt="sampler_type", metavar="SAMPLER_TYPE", default=fs.sampler_type,
                         help="use SAMPLER_TYPE (z48/mpc4k/s56k|mock_z48) [default: %default]")
    fs.parser.add_option(mountopt="connector", metavar="CONNECTOR", default=fs.connector,
                         help="use CONNECTOR (usb|osc) [default: %default]")
    fs.parser.add_option(mountopt="sample_file", metavar="SAMPLE_FILE", 
                         help="The wave file to be served when any wave file is accessed on the filesystem (only valid with mock_z48)")
    fs.parser.add_option(mountopt="program_file", metavar="PROGRAM_FILE", 
                         help="The program to be served when any program is accessed on the filesystem (only valid with mock_z48")

    fs.parse(values=fs, errex=1)

    if fs.sampler_type == "mock_z48":
        sampler = Devices.get_instance('mock_z48', 'mock')
        sampler.set_sample(fs.sample_file)
        sampler.set_program(fs.program_file)
    else:
        sampler = Devices.get_instance(fs.sampler_type, fs.connector)
    try:
        #sampler.start_osc_server()
        fs.main(sampler)
    finally:
        #sampler.stop_osc_server()
        sampler.close()




""" Python equivalent of akai section programtools

Methods to manipulate sampler programs
"""

__author__ =  'Walco van Loon'
__version__ =  '0.2'

from aksy.devices.akai.sysex import Command

import aksy.devices.akai.sysex_types

class Programtools:
    def __init__(self, z48):
        self.sampler = z48
        self.get_no_items_cmd = Command('_', '\x14\x01', 'programtools', 'get_no_items', (), None)
        self.get_handles_cmd = Command('_', '\x14\x02\x00', 'programtools', 'get_handles', (), None)
        self.get_names_cmd = Command('_', '\x14\x02\x01', 'programtools', 'get_names', (), None)
        self.get_handles_names_cmd = Command('_', '\x14\x02\x02', 'programtools', 'get_handles_names', (), None)
        self.get_modified_cmd = Command('_', '\x14\x02\x03', 'programtools', 'get_modified', (), None)
        self.set_curr_by_handle_cmd = Command('_', '\x14\x03', 'programtools', 'set_curr_by_handle', (aksy.devices.akai.sysex_types.DWORD,), None)
        self.set_curr_by_name_cmd = Command('_', '\x14\x04', 'programtools', 'set_curr_by_name', (aksy.devices.akai.sysex_types.STRING,), None)
        self.get_curr_handle_cmd = Command('_', '\x14\x05', 'programtools', 'get_curr_handle', (), None)
        self.get_curr_name_cmd = Command('_', '\x14\x06', 'programtools', 'get_curr_name', (), None)
        self.get_name_by_handle_cmd = Command('_', '\x14\x07', 'programtools', 'get_name_by_handle', (aksy.devices.akai.sysex_types.DWORD,), None)
        self.get_handle_by_name_cmd = Command('_', '\x14\x08', 'programtools', 'get_handle_by_name', (aksy.devices.akai.sysex_types.STRING,), None)
        self.delete_all_cmd = Command('_', '\x14\x09', 'programtools', 'delete_all', (), None)
        self.delete_curr_cmd = Command('_', '\x14\x0A', 'programtools', 'delete_curr', (), None)
        self.delete_by_handle_cmd = Command('_', '\x14\x0B', 'programtools', 'delete_by_handle', (aksy.devices.akai.sysex_types.DWORD,), None)
        self.rename_curr_cmd = Command('_', '\x14\x0C', 'programtools', 'rename_curr', (aksy.devices.akai.sysex_types.STRING,), None)
        self.rename_by_handle_cmd = Command('_', '\x14\x0D', 'programtools', 'rename_by_handle', (aksy.devices.akai.sysex_types.DWORD, aksy.devices.akai.sysex_types.STRING), None)
        self.tag_cmd = Command('_', '\x14\x0E', 'programtools', 'tag', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.get_tag_bitmap_cmd = Command('_', '\x14\x0F', 'programtools', 'get_tag_bitmap', (), None)
        self.get_modified_name_cmd = Command('_', '\x14\x10', 'programtools', 'get_modified_name', (), None)
        self.get_modified_state_cmd = Command('_', '\x14\x11', 'programtools', 'get_modified_state', (), None)
        self.delete_tagged_cmd = Command('_', '\x14\x18', 'programtools', 'delete_tagged', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.create_new_cmd = Command('_', '\x14\x40', 'programtools', 'create_new', (aksy.devices.akai.sysex_types.WORD, aksy.devices.akai.sysex_types.STRING), None)
        self.add_keygroups_cmd = Command('_', '\x14\x41', 'programtools', 'add_keygroups', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.delete_keygroup_cmd = Command('_', '\x14\x42', 'programtools', 'delete_keygroup', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.delete_blank_keygroups_cmd = Command('_', '\x14\x43', 'programtools', 'delete_blank_keygroups', (), None)
        self.arrange_keygroups_cmd = Command('_', '\x14\x44', 'programtools', 'arrange_keygroups', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.copy_keygroup_cmd = Command('_', '\x14\x45', 'programtools', 'copy_keygroup', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.copy_cmd = Command('_', '\x14\x48', 'programtools', 'copy', (aksy.devices.akai.sysex_types.STRING,), None)
        self.merge_programs_cmd = Command('_', '\x14\x49', 'programtools', 'merge_programs', (aksy.devices.akai.sysex_types.DWORD, aksy.devices.akai.sysex_types.DWORD), None)
        self.add_keygroup_sample_cmd = Command('_', '\x14\x4A', 'programtools', 'add_keygroup_sample', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BOOL, aksy.devices.akai.sysex_types.STRING), None)
        self.copy_temperament_to_user_cmd = Command('_', '\x14\x50', 'programtools', 'copy_temperament_to_user', (), None)
        self.get_no_modulation_connections_cmd = Command('_', '\x14\x54', 'programtools', 'get_no_modulation_connections', (), None)
        self.get_no_modulation_sources_cmd = Command('_', '\x14\x55', 'programtools', 'get_no_modulation_sources', (), None)
        self.get_no_modulation_destinations_cmd = Command('_', '\x14\x56', 'programtools', 'get_no_modulation_destinations', (), None)
        self.get_name_modulation_source_cmd = Command('_', '\x14\x57', 'programtools', 'get_name_modulation_source', (aksy.devices.akai.sysex_types.WORD,), None)
        self.get_name_modulation_dest_cmd = Command('_', '\x14\x58', 'programtools', 'get_name_modulation_dest', (aksy.devices.akai.sysex_types.WORD,), None)
        self.get_group_id_cmd = Command('_', '\x17\x01', 'programtools', 'get_group_id', (), None)
        self.get_type_cmd = Command('_', '\x17\x03', 'programtools', 'get_type', (), None)
        self.get_genre_cmd = Command('_', '\x17\x04', 'programtools', 'get_genre', (), None)
        self.get_program_no_cmd = Command('_', '\x17\x08', 'programtools', 'get_program_no', (), None)
        self.get_no_keygroups_cmd = Command('_', '\x17\x09', 'programtools', 'get_no_keygroups', (), None)
        self.get_keygroup_xfade_cmd = Command('_', '\x17\x0A', 'programtools', 'get_keygroup_xfade', (), None)
        self.get_keygroup_xfade_type_cmd = Command('_', '\x17\x0B', 'programtools', 'get_keygroup_xfade_type', (), None)
        self.get_level_cmd = Command('_', '\x17\x0C', 'programtools', 'get_level', (), None)
        self.get_polyphony_cmd = Command('_', '\x17\x10', 'programtools', 'get_polyphony', (), None)
        self.get_reassignment_method_cmd = Command('_', '\x17\x11', 'programtools', 'get_reassignment_method', (), None)
        self.get_softpedal_loudness_reduction_cmd = Command('_', '\x17\x12', 'programtools', 'get_softpedal_loudness_reduction', (), None)
        self.get_softpedal_attack_stretch_cmd = Command('_', '\x17\x13', 'programtools', 'get_softpedal_attack_stretch', (), None)
        self.get_softpedal_filter_close_cmd = Command('_', '\x17\x14', 'programtools', 'get_softpedal_filter_close', (), None)
        self.get_midi_transpose_cmd = Command('_', '\x17\x15', 'programtools', 'get_midi_transpose', (), None)
        self.get_mpc_pad_assignment_cmd = Command('_', '\x17\x18', 'programtools', 'get_mpc_pad_assignment', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_modulation_connection_cmd = Command('_', '\x17\x20', 'programtools', 'get_modulation_connection', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.WORD), None)
        self.get_modulation_source_type_cmd = Command('_', '\x17\x21', 'programtools', 'get_modulation_source_type', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_modulation_destination_type_cmd = Command('_', '\x17\x22', 'programtools', 'get_modulation_destination_type', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_modulation_level_cmd = Command('_', '\x17\x23', 'programtools', 'get_modulation_level', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.WORD), None)
        self.get_midi_controller_number_cmd = Command('_', '\x17\x24', 'programtools', 'get_midi_controller_number', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_edit_keygroup_cmd = Command('_', '\x17\x25', 'programtools', 'get_edit_keygroup', (), None)
        self.get_modulation_level_edit_keygroup_cmd = Command('_', '\x17\x26', 'programtools', 'get_modulation_level_edit_keygroup', (), None)
        self.get_tune_cmd = Command('_', '\x17\x30', 'programtools', 'get_tune', (), None)
        self.get_temperament_template_cmd = Command('_', '\x17\x31', 'programtools', 'get_temperament_template', (), None)
        self.get_program_temperament_cmd = Command('_', '\x17\x32', 'programtools', 'get_program_temperament', (), None)
        self.get_key_cmd = Command('_', '\x17\x33', 'programtools', 'get_key', (), None)
        self.get_user_temperament_note_cmd = Command('_', '\x17\x34', 'programtools', 'get_user_temperament_note', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_pitchbend_up_cmd = Command('_', '\x17\x40', 'programtools', 'get_pitchbend_up', (), None)
        self.get_pitchbend_down_cmd = Command('_', '\x17\x41', 'programtools', 'get_pitchbend_down', (), None)
        self.get_pitchbend_mode_cmd = Command('_', '\x17\x42', 'programtools', 'get_pitchbend_mode', (), None)
        self.get_aftertouch_value_cmd = Command('_', '\x17\x43', 'programtools', 'get_aftertouch_value', (), None)
        self.get_legato_cmd = Command('_', '\x17\x44', 'programtools', 'get_legato', (), None)
        self.get_portamento_enabled_cmd = Command('_', '\x17\x45', 'programtools', 'get_portamento_enabled', (), None)
        self.get_portamento_mode_cmd = Command('_', '\x17\x46', 'programtools', 'get_portamento_mode', (), None)
        self.get_portamento_time_cmd = Command('_', '\x17\x47', 'programtools', 'get_portamento_time', (), None)
        self.get_glissando_mode_cmd = Command('_', '\x17\x48', 'programtools', 'get_glissando_mode', (), None)
        self.get_aftertouch_mode_cmd = Command('_', '\x17\x49', 'programtools', 'get_aftertouch_mode', (), None)
        self.set_group_id_cmd = Command('_', '\x16\x01', 'programtools', 'set_group_id', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_type_cmd = Command('_', '\x16\x03', 'programtools', 'set_type', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_genre_cmd = Command('_', '\x16\x04', 'programtools', 'set_genre', (aksy.devices.akai.sysex_types.STRING,), None)
        self.set_program_no_cmd = Command('_', '\x16\x08', 'programtools', 'set_program_no', (aksy.devices.akai.sysex_types.WORD,), None)
        self.set_no_keygroups_cmd = Command('_', '\x16\x09', 'programtools', 'set_no_keygroups', (aksy.devices.akai.sysex_types.WORD,), None)
        self.set_keygroup_xfade_cmd = Command('_', '\x16\x0A', 'programtools', 'set_keygroup_xfade', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_keygroup_xfade_type_cmd = Command('_', '\x16\x0B', 'programtools', 'set_keygroup_xfade_type', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_level_cmd = Command('_', '\x16\x0C', 'programtools', 'set_level', (aksy.devices.akai.sysex_types.SWORD,), None)
        self.set_polyphony_cmd = Command('_', '\x16\x10', 'programtools', 'set_polyphony', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_reassignment_method_cmd = Command('_', '\x16\x11', 'programtools', 'set_reassignment_method', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_softpedal_loudness_reduction_cmd = Command('_', '\x16\x12', 'programtools', 'set_softpedal_loudness_reduction', (), None)
        self.set_softpedal_attack_stretch_cmd = Command('_', '\x16\x13', 'programtools', 'set_softpedal_attack_stretch', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_softpedal_filter_close_cmd = Command('_', '\x16\x14', 'programtools', 'set_softpedal_filter_close', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_midi_transpose_cmd = Command('_', '\x16\x15', 'programtools', 'set_midi_transpose', (aksy.devices.akai.sysex_types.SBYTE,), None)
        self.set_mpc_pad_assignment_cmd = Command('_', '\x16\x18', 'programtools', 'set_mpc_pad_assignment', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.set_modulation_connection_cmd = Command('_', '\x16\x20', 'programtools', 'set_modulation_connection', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.WORD, aksy.devices.akai.sysex_types.WORD, aksy.devices.akai.sysex_types.WORD, aksy.devices.akai.sysex_types.SBYTE), None)
        self.set_modulation_source_cmd = Command('_', '\x16\x21', 'programtools', 'set_modulation_source', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.WORD), None)
        self.set_modulation_destination_cmd = Command('_', '\x16\x22', 'programtools', 'set_modulation_destination', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.WORD), None)
        self.set_modulation_level_cmd = Command('_', '\x16\x23', 'programtools', 'set_modulation_level', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.WORD, aksy.devices.akai.sysex_types.SBYTE), None)
        self.set_midi_ctrl_no_cmd = Command('_', '\x16\x24', 'programtools', 'set_midi_ctrl_no', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.set_edit_keygroup_cmd = Command('_', '\x16\x25', 'programtools', 'set_edit_keygroup', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.WORD), None)
        self.set_edit_kegyroup_modulation_level_cmd = Command('_', '\x16\x26', 'programtools', 'set_edit_kegyroup_modulation_level', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.set_tune_cmd = Command('_', '\x16\x30', 'programtools', 'set_tune', (aksy.devices.akai.sysex_types.TUNE,), None)
        self.set_temperament_template_cmd = Command('_', '\x16\x31', 'programtools', 'set_temperament_template', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_program_temperament_cmd = Command('_', '\x16\x32', 'programtools', 'set_program_temperament', (aksy.devices.akai.sysex_types.SBYTE,), None)
        self.set_key_cmd = Command('_', '\x16\x33', 'programtools', 'set_key', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_user_temperament_note_cmd = Command('_', '\x16\x34', 'programtools', 'set_user_temperament_note', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.SBYTE), None)
        self.set_pitchbend_up_cmd = Command('_', '\x16\x40', 'programtools', 'set_pitchbend_up', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_pitchbend_down_cmd = Command('_', '\x16\x41', 'programtools', 'set_pitchbend_down', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_pitchbend_mode_cmd = Command('_', '\x16\x42', 'programtools', 'set_pitchbend_mode', (aksy.devices.akai.sysex_types.BOOL,), None)
        self.set_aftertouch_value_cmd = Command('_', '\x16\x43', 'programtools', 'set_aftertouch_value', (aksy.devices.akai.sysex_types.SBYTE,), None)
        self.set_legato_cmd = Command('_', '\x16\x44', 'programtools', 'set_legato', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_portamento_enabled_cmd = Command('_', '\x16\x45', 'programtools', 'set_portamento_enabled', (aksy.devices.akai.sysex_types.BOOL,), None)
        self.set_portamento_mode_cmd = Command('_', '\x16\x46', 'programtools', 'set_portamento_mode', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_portamento_time_cmd = Command('_', '\x16\x47', 'programtools', 'set_portamento_time', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_glissando_mode_cmd = Command('_', '\x16\x48', 'programtools', 'set_glissando_mode', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_aftertouch_mode_cmd = Command('_', '\x16\x49', 'programtools', 'set_aftertouch_mode', (aksy.devices.akai.sysex_types.BYTE,), None)

    def get_no_items(self):
        """Get number of items in memory
        """
        return self.sampler.execute(self.get_no_items_cmd, ())

    def get_handles(self):
        """Get list of info for all items: 0=list of handles;

        Returns:
            DWORD
        """
        return self.sampler.execute(self.get_handles_cmd, ())

    def get_names(self):
        """Get list of info for all items: 1=list of names

        Returns:
            STRINGARRAY
        """
        return self.sampler.execute(self.get_names_cmd, ())

    def get_handles_names(self):
        """Get list of info for all items: 2=list of handle+name;

        Returns:
            HANDLENAMEARRAY
        """
        return self.sampler.execute(self.get_handles_names_cmd, ())

    def get_modified(self):
        """Get list of info for all items: 3=list of handle+modified/tagged name

        Returns:
            HANDLENAMEARRAY
        """
        return self.sampler.execute(self.get_modified_cmd, ())

    def set_curr_by_handle(self, arg0):
        """Select current item by handle
        """
        return self.sampler.execute(self.set_curr_by_handle_cmd, (arg0, ))

    def set_curr_by_name(self, arg0):
        """Select current item by name
        """
        return self.sampler.execute(self.set_curr_by_name_cmd, (arg0, ))

    def get_curr_handle(self):
        """Get handle of current item

        Returns:
            DWORD
        """
        return self.sampler.execute(self.get_curr_handle_cmd, ())

    def get_curr_name(self):
        """Get name of current item

        Returns:
            STRING
        """
        return self.sampler.execute(self.get_curr_name_cmd, ())

    def get_name_by_handle(self, arg0):
        """Get item name from handle

        Returns:
            STRING
        """
        return self.sampler.execute(self.get_name_by_handle_cmd, (arg0, ))

    def get_handle_by_name(self, arg0):
        """Get item handle from name

        Returns:
            DWORD
        """
        return self.sampler.execute(self.get_handle_by_name_cmd, (arg0, ))

    def delete_all(self):
        """Delete ALL items from memory
        """
        return self.sampler.execute(self.delete_all_cmd, ())

    def delete_curr(self):
        """Delete current item from memory
        """
        return self.sampler.execute(self.delete_curr_cmd, ())

    def delete_by_handle(self, arg0):
        """Delete item represented by handle <Data1>
        """
        return self.sampler.execute(self.delete_by_handle_cmd, (arg0, ))

    def rename_curr(self, arg0):
        """Rename current item
        """
        return self.sampler.execute(self.rename_curr_cmd, (arg0, ))

    def rename_by_handle(self, arg0, arg1):
        """Rename item represented by handle <Data1>
        """
        return self.sampler.execute(self.rename_by_handle_cmd, (arg0, arg1, ))

    def tag(self, arg0, arg1):
        """Set Tag Bit <Data1> = bit to set(0-7), <Data2> = (0=OFF, 1=ON), Data3> = (0=CURRENT, 1=ALL)
        """
        return self.sampler.execute(self.tag_cmd, (arg0, arg1, ))

    def get_tag_bitmap(self):
        """Get Tag Bitmap

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_tag_bitmap_cmd, ())

    def get_modified_name(self):
        """Get name of current item with modified/tagged info.
        """
        return self.sampler.execute(self.get_modified_name_cmd, ())

    def get_modified_state(self):
        """Get modified state of current item.

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_modified_state_cmd, ())

    def delete_tagged(self, arg0):
        """Delete tagged items <Data1> = tag bit
        """
        return self.sampler.execute(self.delete_tagged_cmd, (arg0, ))

    def create_new(self, arg0, arg1):
        """Create New Program <Data1> = number of keygroups;<Data2> = name.
        """
        return self.sampler.execute(self.create_new_cmd, (arg0, arg1, ))

    def add_keygroups(self, arg0):
        """Add Keygroups to Program <Data1> = number of keygroups to add
        """
        return self.sampler.execute(self.add_keygroups_cmd, (arg0, ))

    def delete_keygroup(self, arg0):
        """Delete Keygroup (keygroup index)
        """
        return self.sampler.execute(self.delete_keygroup_cmd, (arg0, ))

    def delete_blank_keygroups(self):
        """Delete Blank Keygroups
        """
        return self.sampler.execute(self.delete_blank_keygroups_cmd, ())

    def arrange_keygroups(self, arg0):
        """Arrange Keygroups (note 0:orig 1:low 2:high)
        """
        return self.sampler.execute(self.arrange_keygroups_cmd, (arg0, ))

    def copy_keygroup(self, arg0):
        """Copy Keygroup (keygroup  index)
        """
        return self.sampler.execute(self.copy_keygroup_cmd, (arg0, ))

    def copy(self, arg0):
        """Copy Program (program name)
        """
        return self.sampler.execute(self.copy_cmd, (arg0, ))

    def merge_programs(self, arg0, arg1):
        """Merge Programs (program handle1, handle2)
        """
        return self.sampler.execute(self.merge_programs_cmd, (arg0, arg1, ))

    def add_keygroup_sample(self, arg0, arg1, arg2, arg3, arg4):
        """Add Keygroup Sample (low note, high note, zone, keytrack, sample name)
        """
        return self.sampler.execute(self.add_keygroup_sample_cmd, (arg0, arg1, arg2, arg3, arg4, ))

    def copy_temperament_to_user(self):
        """Copies Program Temperament to User Temperament
        """
        return self.sampler.execute(self.copy_temperament_to_user_cmd, ())

    def get_no_modulation_connections(self):
        """Get number of Modulation Connections

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_no_modulation_connections_cmd, ())

    def get_no_modulation_sources(self):
        """Get number of Modulation Sources

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_no_modulation_sources_cmd, ())

    def get_no_modulation_destinations(self):
        """Get number of Modulation Destinations

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_no_modulation_destinations_cmd, ())

    def get_name_modulation_source(self, arg0):
        """Get Name of Modulation Source (source index)

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_name_modulation_source_cmd, (arg0, ))

    def get_name_modulation_dest(self, arg0):
        """Get Name of Modulation Destination (dest index)

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_name_modulation_dest_cmd, (arg0, ))

    def get_group_id(self):
        """Get Group ID

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_group_id_cmd, ())

    def get_type(self):
        """Get Program Type <Reply> = (0=KEYGROUP, 1=DRUM)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_type_cmd, ())

    def get_genre(self):
        """Get Genre

        Returns:
            STRING
        """
        return self.sampler.execute(self.get_genre_cmd, ())

    def get_program_no(self):
        """Get Program Number <Reply1> = (0=OFF, 1=128)

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_program_no_cmd, ())

    def get_no_keygroups(self):
        """Get Number of keygroups

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_no_keygroups_cmd, ())

    def get_keygroup_xfade(self):
        """Get Keygroup Crossfade <Reply1> = (0=OFF, 1=ON)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_keygroup_xfade_cmd, ())

    def get_keygroup_xfade_type(self):
        """Get Keygroup Crossfade type <Reply1> = (0=LIN, 1=EXP, 2=LOG)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_keygroup_xfade_type_cmd, ())

    def get_level(self):
        """Get Program Level <Reply1> = level in 10 dB

        Returns:
            SWORD
        """
        return self.sampler.execute(self.get_level_cmd, ())

    def get_polyphony(self):
        """Get Polyphony

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_polyphony_cmd, ())

    def get_reassignment_method(self):
        """Get Reassignment <Reply1> = (0=QUIETEST, 1=OLDEST)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_reassignment_method_cmd, ())

    def get_softpedal_loudness_reduction(self):
        """Soft Pedal Loudness Reduction

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_softpedal_loudness_reduction_cmd, ())

    def get_softpedal_attack_stretch(self):
        """Soft Pedal Attack Stretch

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_softpedal_attack_stretch_cmd, ())

    def get_softpedal_filter_close(self):
        """Soft Pedal Filter Close

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_softpedal_filter_close_cmd, ())

    def get_midi_transpose(self):
        """Get midi transpose

        Returns:
            SBYTE
        """
        return self.sampler.execute(self.get_midi_transpose_cmd, ())

    def get_mpc_pad_assignment(self, arg0):
        """Get the midi pad assignment (pad index)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_mpc_pad_assignment_cmd, (arg0, ))

    def get_modulation_connection(self, arg0, arg1):
        """Get the modulation connection(pin number, keygroup for level - note 0 or 'all' is not supported)

        Returns:
            WORD
            WORD
            SBYTE
        """
        return self.sampler.execute(self.get_modulation_connection_cmd, (arg0, arg1, ))

    def get_modulation_source_type(self, arg0):
        """Get the modulation source type (pin number)

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_modulation_source_type_cmd, (arg0, ))

    def get_modulation_destination_type(self, arg0):
        """Get the modulation dest type (pin number)

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_modulation_destination_type_cmd, (arg0, ))

    def get_modulation_level(self, arg0, arg1):
        """Get the modulation level (pin number, keygroup number - note 0 or 'all' is not supported)

        Returns:
            SBYTE
        """
        return self.sampler.execute(self.get_modulation_level_cmd, (arg0, arg1, ))

    def get_midi_controller_number(self, arg0):
        """Get the midi controller number (pin number - only available if source=CTRL)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_midi_controller_number_cmd, (arg0, ))

    def get_edit_keygroup(self):
        """Get edit keygroup

        Returns:
            BYTE
            WORD
        """
        return self.sampler.execute(self.get_edit_keygroup_cmd, ())

    def get_modulation_level_edit_keygroup(self):
        """Get Modulation Level of Edit Keygroup

        Returns:
            SBYTE
        """
        return self.sampler.execute(self.get_modulation_level_edit_keygroup_cmd, ())

    def get_tune(self):
        """Get Cents Tune

        Returns:
            TUNE
        """
        return self.sampler.execute(self.get_tune_cmd, ())

    def get_temperament_template(self):
        """Get Temperament Template, where <Reply1> = template (see Table 22)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_temperament_template_cmd, ())

    def get_program_temperament(self):
        """Get Program Temperament

        Returns:
            SBYTE
        """
        return self.sampler.execute(self.get_program_temperament_cmd, ())

    def get_key(self):
        """Get Key

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_key_cmd, ())

    def get_user_temperament_note(self, arg0):
        """Get User Temperament Note

        Returns:
            TUNE
        """
        return self.sampler.execute(self.get_user_temperament_note_cmd, (arg0, ))

    def get_pitchbend_up(self):
        """Get Pitch Bend Up

        Returns:
            BYT
        """
        return self.sampler.execute(self.get_pitchbend_up_cmd, ())

    def get_pitchbend_down(self):
        """Get Pitch Bend down

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_pitchbend_down_cmd, ())

    def get_pitchbend_mode(self):
        """Get Pitch Bend Mode (normal, held)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_pitchbend_mode_cmd, ())

    def get_aftertouch_value(self):
        """Get Aftertouch Value

        Returns:
            SBYTE
        """
        return self.sampler.execute(self.get_aftertouch_value_cmd, ())

    def get_legato(self):
        """Get Legato Setting (off, pitch, loop

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_legato_cmd, ())

    def get_portamento_enabled(self):
        """Get Portamento Enabled

        Returns:
            BOOL
        """
        return self.sampler.execute(self.get_portamento_enabled_cmd, ())

    def get_portamento_mode(self):
        """Get portamento mode (time, rate)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_portamento_mode_cmd, ())

    def get_portamento_time(self):
        """Get portamento time

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_portamento_time_cmd, ())

    def get_glissando_mode(self):
        """Get Glissando Mode

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_glissando_mode_cmd, ())

    def get_aftertouch_mode(self):
        """Get Aftertouch Type

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_aftertouch_mode_cmd, ())

    def set_group_id(self, arg0):
        """Set Group ID
        """
        return self.sampler.execute(self.set_group_id_cmd, (arg0, ))

    def set_type(self, arg0):
        """Set Program Type <Data1> = (0=KEYGROUP, 1=DRUM)
        """
        return self.sampler.execute(self.set_type_cmd, (arg0, ))

    def set_genre(self, arg0):
        """Set Genre
        """
        return self.sampler.execute(self.set_genre_cmd, (arg0, ))

    def set_program_no(self, arg0):
        """Set Program Number <Data1> = (0=OFF, 1=128)
        """
        return self.sampler.execute(self.set_program_no_cmd, (arg0, ))

    def set_no_keygroups(self, arg0):
        """Set Number of keygroups
        """
        return self.sampler.execute(self.set_no_keygroups_cmd, (arg0, ))

    def set_keygroup_xfade(self, arg0):
        """Set Keygroup Crossfade <Data1> = (0=OFF, 1=ON)
        """
        return self.sampler.execute(self.set_keygroup_xfade_cmd, (arg0, ))

    def set_keygroup_xfade_type(self, arg0):
        """Set Keygroup Crossfade type <Data1> = (0=LIN, 1=EXP, 2=LOG)
        """
        return self.sampler.execute(self.set_keygroup_xfade_type_cmd, (arg0, ))

    def set_level(self, arg0):
        """Set Program Level <Data1> = level in 10dB (-600 +60)
        """
        return self.sampler.execute(self.set_level_cmd, (arg0, ))

    def set_polyphony(self, arg0):
        """Set Polyphony
        """
        return self.sampler.execute(self.set_polyphony_cmd, (arg0, ))

    def set_reassignment_method(self, arg0):
        """Set Reassignment <Data1> = (0=QUIETEST, 1=OLDEST)
        """
        return self.sampler.execute(self.set_reassignment_method_cmd, (arg0, ))

    def set_softpedal_loudness_reduction(self):
        """Soft Pedal Loudness Reduction
        """
        return self.sampler.execute(self.set_softpedal_loudness_reduction_cmd, ())

    def set_softpedal_attack_stretch(self, arg0):
        """Soft Pedal Attack Stretch
        """
        return self.sampler.execute(self.set_softpedal_attack_stretch_cmd, (arg0, ))

    def set_softpedal_filter_close(self, arg0):
        """Soft Pedal Filter Close
        """
        return self.sampler.execute(self.set_softpedal_filter_close_cmd, (arg0, ))

    def set_midi_transpose(self, arg0):
        """Midi Transpose (-36 +36)
        """
        return self.sampler.execute(self.set_midi_transpose_cmd, (arg0, ))

    def set_mpc_pad_assignment(self, arg0, arg1):
        """MPC pad assignment <Data1> = pad, <Data2> = note
        """
        return self.sampler.execute(self.set_mpc_pad_assignment_cmd, (arg0, arg1, ))

    def set_modulation_connection(self, arg0, arg1, arg2, arg3, arg4):
        """Set Modulation Connection <Data1> = connection (pin) number;<Data2> = keygroup number (0=ALL, 1-128=KEYGROUP) <Data3> = source (see Table 24); <Data4> = destination (see Table 25); <Data5> = level.  If Source or Destination is zero, the connection will be cleared.
        """
        return self.sampler.execute(self.set_modulation_connection_cmd, (arg0, arg1, arg2, arg3, arg4, ))

    def set_modulation_source(self, arg0, arg1):
        """Set Modulation Source (see Table 24)
        """
        return self.sampler.execute(self.set_modulation_source_cmd, (arg0, arg1, ))

    def set_modulation_destination(self, arg0, arg1):
        """Set Modulation Destination (see Table 25)
        """
        return self.sampler.execute(self.set_modulation_destination_cmd, (arg0, arg1, ))

    def set_modulation_level(self, arg0, arg1, arg2):
        """Set Modulation Level <Data1> = pin number; <Data2> = (0=ALL, 1-128=KEYGROUP); <Data3> = level
        """
        return self.sampler.execute(self.set_modulation_level_cmd, (arg0, arg1, arg2, ))

    def set_midi_ctrl_no(self, arg0, arg1):
        """Set MIDI controller number (only used if Source = CTRL)
        """
        return self.sampler.execute(self.set_midi_ctrl_no_cmd, (arg0, arg1, ))

    def set_edit_keygroup(self, arg0, arg1):
        """Set Edit Keygroup (used to edit level) <Data2> = Edit Keygroup
        """
        return self.sampler.execute(self.set_edit_keygroup_cmd, (arg0, arg1, ))

    def set_edit_kegyroup_modulation_level(self, arg0, arg1):
        """Set Modulation Level of Edit Keygroup
        """
        return self.sampler.execute(self.set_edit_kegyroup_modulation_level_cmd, (arg0, arg1, ))

    def set_tune(self, arg0):
        """Set Cents Tune
        """
        return self.sampler.execute(self.set_tune_cmd, (arg0, ))

    def set_temperament_template(self, arg0):
        """Set Temperament Template, where <Data1> = template (see Table 22)
        """
        return self.sampler.execute(self.set_temperament_template_cmd, (arg0, ))

    def set_program_temperament(self, arg0):
        """Set Program Temperament 0=C, 1=C#, 2=D, 3=Eb, 4=E, 5=F, 6=F#, 7=G, 8=G#, 9=A, 10=Bb, 11=B
        """
        return self.sampler.execute(self.set_program_temperament_cmd, (arg0, ))

    def set_key(self, arg0):
        """Set Key 0=C, 1=C#, 2=D, 3=Eb, 4=E, 5=F, 6=F#, 7=G, 8=G#, 9=A, 10=Bb, 11=B
        """
        return self.sampler.execute(self.set_key_cmd, (arg0, ))

    def set_user_temperament_note(self, arg0, arg1):
        """Set User Temperament Note (note, cents)
        """
        return self.sampler.execute(self.set_user_temperament_note_cmd, (arg0, arg1, ))

    def set_pitchbend_up(self, arg0):
        """Set pitchbend up (semitones)
        """
        return self.sampler.execute(self.set_pitchbend_up_cmd, (arg0, ))

    def set_pitchbend_down(self, arg0):
        """Set pitchbend down (semitones)
        """
        return self.sampler.execute(self.set_pitchbend_down_cmd, (arg0, ))

    def set_pitchbend_mode(self, arg0):
        """Set bend mode (held=true)
        """
        return self.sampler.execute(self.set_pitchbend_mode_cmd, (arg0, ))

    def set_aftertouch_value(self, arg0):
        """Set aftertouch value
        """
        return self.sampler.execute(self.set_aftertouch_value_cmd, (arg0, ))

    def set_legato(self, arg0):
        """Set legato setting (off, pitch, loop)
        """
        return self.sampler.execute(self.set_legato_cmd, (arg0, ))

    def set_portamento_enabled(self, arg0):
        """Enable portamento
        """
        return self.sampler.execute(self.set_portamento_enabled_cmd, (arg0, ))

    def set_portamento_mode(self, arg0):
        """Set portamento mode (time, rate)
        """
        return self.sampler.execute(self.set_portamento_mode_cmd, (arg0, ))

    def set_portamento_time(self, arg0):
        """Set portamento time
        """
        return self.sampler.execute(self.set_portamento_time_cmd, (arg0, ))

    def set_glissando_mode(self, arg0):
        """Set glissando mode (portamento, glissando)
        """
        return self.sampler.execute(self.set_glissando_mode_cmd, (arg0, ))

    def set_aftertouch_mode(self, arg0):
        """Set aftertouch (channel, poly)
        """
        return self.sampler.execute(self.set_aftertouch_mode_cmd, (arg0, ))




import logging

import sysextools, disktools, programtools, multitools, songtools, multifxtools
import sampletools, systemtools, recordingtools, keygrouptools, zonetools, frontpaneltools
from aksy.devices.akai.sampler import Sampler
from aksy.devices.akai import model

log = logging.getLogger("aksy")

class Z48(Sampler):
    """Z48
    """
    def __init__(self, connector):
        Sampler.__init__(self, connector)
        self.setup_tools()

        self.sysextools.enable_msg_notification(False)
        self.sysextools.enable_item_sync(False)
        
        self.setup_model()

    def setup_tools(self):
        self.disktools = disktools.Disktools(self.connector)
        self.programtools = programtools.Programtools(self.connector)
        self.keygrouptools = keygrouptools.Keygrouptools(self.connector)
        self.zonetools = zonetools.Zonetools(self.connector)
        self.sampletools = sampletools.Sampletools(self.connector)
        self.songtools = songtools.Songtools(self.connector)
        self.multitools = multitools.Multitools(self.connector)
        self.multifxtools = multifxtools.Multifxtools(self.connector)
        self.systemtools = systemtools.Systemtools(self.connector)
        self.sysextools = sysextools.Sysextools(self.connector)
        self.recordingtools = recordingtools.Recordingtools(self.connector)
        self.frontpaneltools = frontpaneltools.Frontpaneltools(self.connector)

    def setup_model(self):
        model.register_handlers({model.Disk: self.disktools,
                        model.FileRef: self.disktools,
                        model.Program: self.programtools,
                        model.Sample: self.sampletools,
                        model.Multi: self.multitools,
                        model.Song: self.songtools})

        self.disks = model.RootDisk('disks', self.disktools.get_disklist())
        self.memory = model.Memory('memory')



import gtk

import panelbase, editors

class KeygroupEnvelopes(panelbase.PanelBase):
    def __init__(self, keygroup, cb):
        self.env_labels = ["Amp Anvelope", "Filter Envelope", "Aux Envelope"]
        panelbase.PanelBase.__init__(self, keygroup, "Envelopes", cb)
    
    def setup(self, keygroup):
        self.clear_children(True)
        self.vbox = gtk.VBox(False, 0)
        self.s = keygroup.s
        self.keygroup = keygroup
        self.update_env('ampenv', self.keygroup.amp_envelope, 0)
        self.update_env('filtenv', self.keygroup.filter_envelope, 1)
        self.update_env('auxenv', self.keygroup.aux_envelope, 2)
        self.pack_start(self.vbox)
        self.show_all()
            
    def update_env(self, envname, env, index):
        setattr(self, envname, editors.EnvelopeHBox(self.keygroup, index))
        label_hbox = gtk.HBox(False,0)
        lb = gtk.Label("%s" % self.env_labels[index])
        lb.set_justify(gtk.JUSTIFY_LEFT)
        label_hbox.pack_start(lb,False,False,5)
        self.vbox.pack_start(label_hbox,False,False,5);
        self.vbox.pack_start(getattr(self, envname),False,False,1);
        e = getattr(self, envname)
        e.show_all()




import samplerobject

class LFO(samplerobject.SamplerObject):
    def __init__(self, s, kg, index):
        samplerobject.SamplerObject.__init__(self, s, kg, "keygrouptools", index)
        self.attrs = ['lfo_rate','lfo_delay','lfo_depth','lfo_waveform','lfo_phase','lfo_shift','lfo_midi_sync','midi_clock_sync_div','lfo_retrigger','lfo_sync']
        self.abbr = {
                "lfo_rate" : "rate",
                "lfo_delay" : "delay",
                "lfo_depth" : "depth",
                "lfo_waveform" : "wave",
                "lfo_phase" : "phase",
                "lfo_shift" : "shift",
                "lfo_midi_sync" : "midi sync",
                "midi_clock_sync_div" : "midi sync div", # ???
                "lfo_retrigger" : "retrig",
                "lfo_sync" : "sync",
                }
        self.keygroup_index = kg.index
        self.need_index_for_set = True
        self.need_index_in_arguments = True
        self.set_current_before_get_set = True
        self.keygroup = kg
        
        self.precache()
        
    def get_handle(self):
        return self.keygroup.p.handle



import gobject

import base

"""
Recording interface(s)!

GOOD IDEA: Put a 'rec' button next to the ... in zone, popup a dialog, record stuff, hit 'OK' and the name of the new sample is returned, select in zone list. Maybe its possible to do this even while playing back??
GOOD IDEA: Chromatic recording of samples from input, use MIDI out (play command should be available to us), automated start/stop, create program, assign samples to zones. Allow spacing of notes and take that into consideration when assigning.
General good idea: base class for these user interfaces to get rid of redundancies and facilitate new interfaces...
"""

class RecordDialog(base.Base):
    def __init__(self, record):
        base.Base.__init__(self, record, "windowRecording")
        self.progressbar = self.xml.get_widget('progressbar_recording')
        self.name = self.xml.get_widget('entry_name')
        self.record = record
        self.update()
        
    def check_progress(self, *args):
        tools = self.record.gettools()
        status = tools.get_status()

        self.progressbar.set_fraction(float(tools.get_progress())/100)
        if status == 6:
            self.continue_check_progress = False
            self.record.gettools().keep()
            print "Kept sample!"
            print self.record.gettools().get_name()

        status = self.record.gettools().get_status()

        if status == 1:
            self.progressbar.set_fraction(0.0)
            self.name.set_text(self.record.gettools().get_name())

        return self.continue_check_progress

    def on_button_clicked(self, widget):
        if widget.name == "button_arm":
            print "before armed", self.record.gettools().get_status()
            self.record.gettools().arm()
            print "after armed", self.record.gettools().get_status()
            if self.record.gettools().get_status() == 2:
                self.record.gettools().record()
            else:
                print self.record.gettools().get_status()
            # 1 == ready
            # 2 == armed
            self.continue_check_progress = True
            gobject.timeout_add(1000, self.check_progress)
        elif widget.name == "button_stop":
            self.record.gettools().stop()
            self.record.gettools().keep()
            self.continue_check_progress = False
            self.progressbar.set_fraction(0.0)
            self.name.set_text(self.record.gettools().get_name())
        elif widget.name == "button_keep":
            self.record.gettools().keep()
            self.continue_check_progress = False
            self.progressbar.set_fraction(0.0)
        elif widget.name == "button_play_preview":
            self.record.gettools().start_playing()
            print self.record.gettools().get_status()
        elif widget.name == "button_stop_preview":
            print "stopping"
            self.record.gettools().stop_playing()
            print self.record.gettools().get_status()



from aksy.devices.akai import sysex

from aksyx import AkaiSampler

import logging

LOG = logging.getLogger("aksy.devices.akai.connector")

class USBConnector(AkaiSampler):
    """ USB Connector for Akai Samplers.
    """
    def __init__(self, device_id):
        AkaiSampler.__init__(self, getattr(USBConnector, device_id.upper()))
        
        #self.sysextools.enable_msg_notification(False)
        #self.sysextools.enable_item_sync(False)

    def execute(self, command, args):
        request = sysex.Request(command, args)
        result_bytes = self.execute_request(request)
        result = sysex.Reply(result_bytes, request.command)
        return result.get_return_value()

    def execute_alt_request(self, handle, commands, args, index = None):
        result_bytes = self.execute_request(sysex.AlternativeRequest(handle, commands, args, index))
        result = sysex.Reply(result_bytes, commands[0], True)
        return result.get_return_value()

    def execute_request(self, request):
        """Execute a request on the sampler.
        Returns the byte response.
        """
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug("Request: %s", repr(request))

        result_bytes = self._execute(request.get_bytes())
        
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug("Response: %s", (sysex.byte_repr(result_bytes)))
        return result_bytes



from aksui.utils import modelutils
import program

class Programs:
    programtypes = {0:"Keygroup", 1:"Drum"}

    def __init__(self,s):
        self.s = s
        pt = s.programtools
        self.no_items = pt.get_no_items()
        self.handles_names = pt.get_handles_names()
        self.names = pt.get_names()

        self.curr_handle = pt.get_curr_handle()
        if self.curr_handle != 268435455:
            self.curr_name = pt.get_curr_name()

        self.programdict = { }        

        i = 0

        while i < len(self.handles_names):
            self.programdict[self.handles_names[i]] = self.handles_names[i+1]
            i += 2

        self.programsmodel = modelutils.get_model_from_list(self.programdict)
        self.programtypesmodel = modelutils.get_model_from_list(Programs.programtypes)

    def getCurr(self):
        if self.curr_handle in self.programdict:
            return program.Program(self.s, self.programdict[self.curr_handle])
        else:
            return None
    
    def getProgram(self, name):
        return program.Program(self.s, name)




import gtk

# from aksui.utils.modelutils import *

LCD_PIXELS_WIDTH = 248
LCD_PIXELS_HEIGHT = 60
BORDER_PIXELS = 2
WINDOW_WIDTH = LCD_PIXELS_WIDTH * 2 + BORDER_PIXELS
WINDOW_HEIGHT = LCD_PIXELS_HEIGHT * 2 + BORDER_PIXELS

w,h = 8,256
mapmap = {}

for x in range(0,256):
    ent = [0,0,0,0,0,0,0,0]
    xx = 0
    for p in [128,64,32,16,8,4,2,1]:
        if (x & p > 0):
            ent[xx] = 1 
        else:
            ent[xx] = 0

        xx = xx + 1

    mapmap[x] = ent

non_ascii_keymap = {"F1":1,"F2":2,"F3":3,"F4":4,"F5":5,"F6":6,"equal":25,"plus":25,"minus":26,"Tab":23,"shift":64}
ascii_keymap = {'BackSpace':8,'Delete':127, 'Insert':272, 'Home':288, 'End':289, 'Page_Up':290, 'Page_Down':291, 'Escape':27, 'grave':774, 'space':32, 'Shift_L':769, 'Shift_R':769, 'Return':13, 'Control_L':770, 'Control_R':770,'period':46,"Left":257,"Right":258,"Up":259,"Down":260,}

class LCDScreen(gtk.DrawingArea):
    def __init__(self,s):
        gtk.DrawingArea.__init__(self)
        self.set_size_request(WINDOW_WIDTH, WINDOW_HEIGHT)

        self.s = s

        self.bv = ''

        self.connect("expose_event", self.on_expose)
        self.connect("realize", self.on_realize)
        self.connect("button_press_event",self.buttonPressEvent)
        self.connect("key_press_event",self.on_key_press_event)
        self.connect("key_release_event",self.on_key_release_event)

        self.add_events(gtk.gdk.KEY_PRESS_MASK|
                        gtk.gdk.KEY_RELEASE_MASK|
                        gtk.gdk.BUTTON_PRESS_MASK |
                        gtk.gdk.BUTTON_RELEASE_MASK |
                        gtk.gdk.POINTER_MOTION_MASK)

        self.set_flags(gtk.CAN_FOCUS)

    def updateLCD(self):
        if self.s:
            pixel_data, control_data = self.s.frontpaneltools.get_panel_state()

            xx,yy = 0,0
            ugiv = {}

            for x in pixel_data:
                x = ord(x)
                for c in mapmap[x]:
                    ugiv[xx] = c 
                    xx = xx + 1

            self.bv = ugiv

    def on_key_release_event(self, widget, event):
        self.handle_keys(widget,event,0)
        self.grab_focus()
        
    def on_key_press_event(self, widget, event):
        self.handle_keys(widget,event,1)
        self.grab_focus()

    def handle_keys(self, widget, event, onoff):
        # We want to ignore irrelevant modifiers like ScrollLock
        # window, multi, fx, edit sample, edit program, record, utilities, save, load
        ctrlmap = { 'w':7,'F1':23,'F2':19,'F3':22,'F4':18,'F5':21,'F6':17,'F7':20,'F8':16 }
        wheelmap = { 'F11':-1, 'F12':1 }
        ignore = ['Control_L','Control_R']

        modifiers = { gtk.gdk.SHIFT_MASK:1 }

        keyname = gtk.gdk.keyval_name(event.keyval)

        if keyname in ignore:
            return

        m = 0

        for mod in modifiers.keys():
            if event.keyval & mod:
                m += modifiers[mod]

        if len(keyname) == 1:
            ascii = ord(keyname)
        else:
            ascii = 0

        fp = self.s.frontpaneltools

        hfunc = None

        if event.state & gtk.gdk.CONTROL_MASK and keyname in ctrlmap.keys():
            funcmap = {1:fp.keypress_hold, 0:fp.keypress_release}
            hfunc = funcmap[onoff]
            arg0 = ctrlmap[keyname]
        elif keyname in wheelmap.keys() and onoff:
            hfunc = fp.move_datawheel
            arg0 = wheelmap[keyname]
        elif keyname in non_ascii_keymap.keys():
            funcmap = {1:fp.keypress_hold, 0:fp.keypress_release}
            hfunc = funcmap[onoff]
            arg0 = non_ascii_keymap[keyname]
        elif ascii > 0 and ascii < 127:
            funcmap = {1:fp.ascii_keypress_hold, 0:fp.ascii_keypress_release}
            hfunc = funcmap[onoff]
            arg0 = ascii
            if onoff:
                arg1 = m
            else:
                arg1 = 0
        elif keyname in ascii_keymap.keys():
            funcmap = {1:fp.ascii_keypress_hold, 0:fp.ascii_keypress_release}
            hfunc = funcmap[onoff]
            arg0 = ascii_keymap[keyname]
            if onoff:
                arg1 = m
            else:
                arg1 = 0

        if hfunc:
            if hfunc.func_code.co_argcount == 3:
                hfunc(arg0, arg1)
            elif hfunc.func_code.co_argcount == 2:
                hfunc(arg0)

            self.updateLCD()
            self.redraw_canvas()


    def buttonPressEvent(self, widget, event):
        x = event.x/2
        y = event.y/2

        if event.type == gtk.gdk.BUTTON_PRESS:
            self.s.frontpaneltools.mouseclick_at_screen(int(x),int(y))
        elif event.type == gtk.gdk._2BUTTON_PRESS:
            self.s.frontpaneltools.mousedoubleclick_at_screen(int(x),int(y))

        self.updateLCD()
        self.redraw_canvas()

        self.grab_focus()

    def redraw_canvas(self):
        rect = gtk.gdk.Rectangle(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT)
        self.window.invalidate_rect(rect, True)
        self.window.process_updates(True)

    def on_realize(self, widget):
        self.gc = widget.window.new_gc()

    def on_expose(self, widget, event):
        self.updateLCD()

        on_color = gtk.gdk.Color(0xFF,0xFF,0xFF,0)
        off_color = gtk.gdk.Color(0,0,0,1)

        if self.bv:
            for y in range(0, LCD_PIXELS_HEIGHT):
                for x in range(0, LCD_PIXELS_WIDTH):
                    val = self.bv[x + (y * LCD_PIXELS_WIDTH)] 
                    xx = x * 2
                    yy = y * 2

                    if val == 1:
                        self.window.draw_point(self.gc,xx,yy)
                        self.window.draw_point(self.gc,xx+1,yy+1)
                        self.window.draw_point(self.gc,xx,yy+1)
                        self.window.draw_point(self.gc,xx+1,yy)

        return False



from StringIO import StringIO
import struct, os.path, os

"""
Reading/writing AKP files

Based on http://mda.smartelectronix.com/akai/AKPspec.html
"""

WAVE_FORMAT_PCM = 0x0001

class Sample:
    """ Models a S5/6000/Z series sample
    which is an ordinary wave file with sampler chunks.
    -dwords manufacturer, product, period
    -midi note
    -pitch fraction
    -smtp format, offset, loopcount
    -loop defs 24 bytes (type, start, end, fraction, count)
    -s5000/6000 specific data:

       Length   Description           Default
       ----------------------------------------------------------------------
          2                           1,0
          1     Original root note    60
          1     Semitone tune +/-36   0
          1     Fine tune     +/-50   0
          1                           0     
          1     Play mode             0=NO LOOPING 1=ONE SHOT 2=LOOP IN REL 3=LOOP TIL REL
          3                           0,0,0
          4     Start point           0
          4     End point             (number of sample words)

    TODO: region support
    """
    def __init__(self, name, pitch, framerate, loops, data):
        self.name = name
        self.pitch = pitch 
        self.framerate = framerate
        self.loops = loops 
        self._sampwidth = 2
        self.nchannels = 1 
        self._nframes = len(data) / (self.nchannels * self._sampwidth)
        self._data = data

    def writefile(self):
        # based on code from the python wave module
        file = open(self.name + '.wav', 'wb')
        file.write('RIFF')
        self._datalength = self._nframes * self.nchannels * self._sampwidth
        data =  struct.pack('<l4s4slhhllhh',
            0, 'WAVE', 'fmt ', 16,
            WAVE_FORMAT_PCM, self.nchannels, self.framerate,
            self.nchannels * self.framerate * self._sampwidth,
            self.nchannels * self._sampwidth, self._sampwidth*8)
         
        file.write(data)
        file.write(self.get_sampler_chunk())
        data = struct.pack('<4sl', 'data', self._datalength)
        file.write(data)
        file.write(self._data)
        length = file.tell()
        file.seek(4)
        file.write(struct.pack('<l', length-8))
        file.close()

    def get_sampler_chunk(self):
        chunk = StringIO() 
            # fraction, smtpe format, offset, count, sampledata
        
        data = struct.pack( '<4s10l', 'smpl', 36 + 18 + 24 * len(self.loops), 
            71, 94, 1000000000/self.framerate, self.pitch, 0, 0, 0, len(self.loops), 18)
        chunk.write(data)
        for cue_id, loop, in enumerate(self.loops):
            chunk.write(loop.get_riff_wavechunk(cue_id))
        
        looptype = len(self.loops) > 0 and self.loops[0].type or 0
        loopstart = len(self.loops) > 0 and self.loops[0].start or 0
        loopend = len(self.loops) > 0 and self.loops[0].end or 0
        # remaining Sampler data
        # unkn, unkn, midi root, semitone t, fine t, unkn, playmode, unkn*3, loopstart, loopend
        chunk.write(
            struct.pack('<10Bll', 2, 0, self.pitch, 0, 0, 0, looptype, 0, 0, 0, loopstart, loopend))
        data =  chunk.getvalue()
        return data
        
class Loop:
    def __init__(self, type, start, end, fraction):
        self.start = start
        self.end = end 
        # 0: no looping, 1: looping 2: one shot
        self.type = type or 0
        self.fraction = 0
        self.count = 0

    def get_riff_wavechunk(self, cue_id):
        """ Returns the 24 byte loop chunk for the wave file
        """
        return struct.pack('<6l', cue_id, 0, self.start, self.end,
                           self.fraction, self.count)

class Base:
    def __init__(self, chunks={}, **kwargs):
        self._chunks = chunks
        self._chunk_index = 0
        if chunks is not None:
            self.setvalues()
            #self.setdefaults()
        else:
            self.setdefaults()
            for kw, value in kwargs.iteritems():
                if not hasattr(self, kw):
                    raise Exception("Unknown property: %s" % kw)
                setattr(self, kw, value)

    def curr_index(self, index, update=1):
        if index is None: 
            index = self._chunk_index
        self._chunk_index += update 
        return index 

    def read_chunk(self, fh, index=None):
        #
        if index is None:
            index = self.curr_index(index)
        fh.seek(index)
        
        bytes_read = fh.read(4)
        assert len(bytes_read) > 0
        
        length = struct.unpack('<l', bytes_read)[0]

        return length, fh.read(length) 

    def read_byte(self, index=None):
        index = self.curr_index(index) 
        return struct.unpack('<b', self._chunk[index])

    def read_string(self, chunk=None, index=None, length=None):
        if not chunk:
            chunk = self._chunk
        index = self.curr_index(index) 
        if length is None:
            length = struct.unpack('<b', chunk[index])[0]
            index = self.curr_index(index) 

        self.curr_index(index, length) 
        string = struct.unpack( '%is' %length, chunk[index: index+length])

class Modulation(Base):  
    """
    Contains modulation routings
    >>> mod = Modulation()
    >>> len(mod.create_chunk())
    46
    """

    NO_SOURCE = 0  
    MODWHEEL = 1
    BEND = 2 
    AFTERTOUCH = 3
    EXTERNAL = 4
    VELOCITY = 5
    KEYBOARD = 6 
    LFO1 = 7
    LFO2 = 8
    AMP_ENV = 9
    FILT_ENV = 10
    AUX_ENV = 11  
    MIDI_MODWHEEL = 12
    MIDI_BEND = 13
    MIDI_EXTERNAL = 14 
    
    def setdefaults(self):
        self.amp_src1 = 6
        self.amp_src2 = 3
        self.pan_src1 = 8
        self.pan_src2 = 6
        self.pan_src3 = 1
        self.lfo1_rate_src = 6
        self.lfo1_delay_src = 6
        self.lfo1_depth_src = 6
        self.lfo2_rate_src = 0
        self.lfo2_delay_src = 0
        self.lfo2_depth_src = 0
        # keygroup mod
        self.pitch1_src = 7
        self.pitch2_src = 11 
        self.amp_src = 11 
        self.filter_input1 = 5 
        self.filter_input2 = 8 
        self.filter_input3 = 9 

    def create_chunk(self):
        return struct.pack('<4sl38b',
            'mods',
            38,
            0,
            0,
            0,
            0,
            0,
            self.amp_src1,
            0,
            self.amp_src2,
            0,
            self.pan_src1,
            0,
            self.pan_src2,
            0,
            self.pan_src3,
            0,
            self.lfo1_rate_src,
            0,
            self.lfo1_delay_src,
            0,
            self.lfo1_depth_src,
            0,
            self.lfo2_rate_src,
            0,
            self.lfo2_delay_src,
            0,
            self.lfo2_depth_src,
            # keygroup mod
            0,
            self.pitch1_src,
            0,
            self.pitch1_src,
            0,
            self.amp_src,
            0,
            self.filter_input1,
            0,
            self.filter_input2,
            0,
            self.filter_input3)

       
NOTE_NAMES = 'a','b', 'bes','c', 'cis', 'd' 'es', 'e', 'f', 'fis', 'g', 'gis'

class Program(Base):
    """
    >>> p = Program('sooperstrings.akp', 'w')
    >>> len(p.create_chunk())
    164
    
    >>> p.midi_prog_no = 1
    >>> p.lfos[0].waveform = LFO.SAW_BI 

    Note the lack of file extension with sample assigning:
    >>> p.keygroups.append(Keygroup())
    >>> p.keygroups[0].low_note = 25
    >>> p.keygroups[0].high_note = 126 
    >>> p.keygroups[0].zones[0].samplename = 'FSM Strings4 C1'
    >>> p.keygroups.append(Keygroup())
    >>> p.keygroups[1].zones[0].samplename = 'test'
    >>> p.keygroups[0].zones[0].low_vel = 33
    >>> p.keygroups[0].zones[0].high_vel = 35 
    >>> p.writefile()
    >>> rp = Program('smstrings.akp', 'r')

    """
    def __init__(self, filename=None, mode='r', **kwargs):
        self.filename = filename
        self._chunks = {}
        chunk = None
        if filename is not None and not os.path.exists(filename):
            raise IOError("File not found: " + filename)
        
        self.fh = file(filename, mode)
        self._read_chunks(self.get_file_length())
                # need more subchunks here
        Base.__init__(self, self._chunks, **kwargs)
        self.keygroups = []

    def get_file_length(self):
        return os.stat(self.filename).st_size
        
    def __del__(self):
        self.fh.close()

    def _read_chunks(self, filelength):
        # collects all chunks adding them to a dict with key chunk label,
        # data value. label and data length bytes are stripped.
        # first bytes are not according to standard, so start with prg chunk
        # prg
        offset = 0x15
        self.fh.seek(0, 1)
        length, self._chunks["prg "] = self.read_chunk(self.fh, offset)
        # prg length
        offset += length
        self._chunk = self._chunks["prg "]
        while offset < filelength:
            length = self.add_chunk(offset)
            if length == 0: break
            offset += length
        
    def add_chunk(self, offset):
        length, chunk = self.read_chunk(self.fh, offset)
        chunk_label = self.read_string(chunk, offset)
        self._chunks[chunk_label] = chunk[offset+4:]
        return length

    def setvalues(self):
        label = self.read_string(index=0x0, length=4)
        print repr(self._chunk)
        assert label == "APRG"
        assert self.read_string(0x5, 4) == "prg "
        self.midi_prog_no = self.read_byte(0x15) 
        self.no_keygrps = self.read_byte(0x16)
        self.loudness = self.read_byte(0x23)
        self.amp_mod1 = self.read_byte()
        self.amp_mod2 = self.read_byte()
        self.pan_mod1 = self.read_byte()
        self.pan_mod2 = self.read_byte()
        self.pan_mod3 = self.read_byte()
        self.velo_sens = self.read_byte()
        self.semi = self.read_byte(0x33)
        self.fine = self.read_byte()
        self.a_detune = self.read_byte()
        self.ais_detune = self.read_byte()
        self.b_detune = self.read_byte()
        self.bes_detune = self.read_byte()
        self.c_detune = self.read_byte()
        self.cis_detune = self.read_byte()
        self.des_detune = self.read_byte()
        self.e_detune = self.read_byte()
        self.f_detune = self.read_byte()
        self.fis_detune = self.read_byte()
        self.g_detune = self.read_byte()
        self.gis_detune = self.read_byte()
        self.pitchbend_up = self.read_byte()
        self.pitchbend_down = self.read_byte()
        self.bend_mode = self.read_byte()
        self.aftertouch = self.read_byte()
        self.lfos = (LFO1(chunk=self._chunk[0x4a:]), 
                     LFO2(chunk=self._chunk[0x4a:]),)
    
    def setdefaults(self):
        self.keygroups = [Keygroup()]
        self.midi_prog_no = 0
        self.no_keygrps = 1 
        self.loudness = 85
        self.amp_mod1 = 0
        self.amp_mod2 = 0
        self.pan_mod1 = 0
        self.pan_mod2 = 0
        self.pan_mod3 = 0
        self.velo_sens = 25
        self.semi = 0 
        self.fine = 0 
        self.a_detune = 0
        self.ais_detune = 0
        self.b_detune = 0
        self.bes_detune = 0
        self.c_detune = 0
        self.cis_detune = 0
        self.des_detune = 0
        self.e_detune = 0
        self.f_detune = 0
        self.fis_detune = 0
        self.g_detune = 0
        self.gis_detune = 0
        self.pitchbend_up = 2
        self.pitchbend_down = 2
        self.bend_mode = 0 # 1: held
        self.aftertouch = 0
        self.lfos = (LFO1(), LFO2(),)
        self.mod = Modulation()

    def create_chunk(self):
        chunk = StringIO()
        # explicitly setting little endian 
        # native mode aligns the longs. eek!
        chunk.write(struct.pack('<4sl8sl6b 4sl8b 4sl24b',
            'RIFF',
            0,
            'APRG'
            'prg ',
            6, 
            1, # set to zero program was not read...
            self.midi_prog_no,
            len(self.keygroups),
            0, 
            0,
            0,
            'out ',
            8,
            0,
            self.loudness,
            self.amp_mod1,
            self.amp_mod2,
            self.pan_mod1,
            self.pan_mod2,
            self.pan_mod3,
            self.velo_sens,
            'tune',
            24,
            0,
            self.semi,
            self.fine,
            self.a_detune,
            self.ais_detune,
            self.b_detune,
            self.bes_detune,
            self.c_detune,
            self.cis_detune,
            self.des_detune,
            self.e_detune,
            self.f_detune,
            self.fis_detune,
            self.g_detune,
            self.gis_detune,
            self.pitchbend_up,
            self.pitchbend_down,
            self.bend_mode,
            self.aftertouch,
            0,
            0,
            0,
            0,
            0))
        chunk.write(''.join([lfo.create_chunk() for lfo in self.lfos]))
        chunk.write(self.mod.create_chunk())
        return chunk.getvalue()

    def writefile(self):
        file = open(self.filename, 'wb')
        file.write(self.create_chunk())
        for kg in self.keygroups:
            file.write(kg.create_chunk())
        file.close()

    def __repr__(self):
        string_repr = StringIO()
        string_repr.write('<Akai Program')
        string_repr.write(''.join(['property: %s, val %s\n' % (item, val) 
                                for item,val in self.__dict__.iteritems()]))
        string_repr.write('>')
        return string_repr.getvalue()

class LFO(Base):
    """ 
    >>> lfo = LFO(waveform=LFO.SAW_BI, rate_mod=36)
    """
    SINE = 0  
    TRIANGLE = 1
    SQUARE = 2
    SQUARE_PLUS = 3
    SQUARE_MINUS = 4
    SAW_BI = 5
    SAW_UP = 6
    SAW_DOWN = 7
    RANDOM = 8

    def setdefaults(self):
        self.waveform = LFO.TRIANGLE 
        self.rate = 43
        self.delay = 0
        self.depth = 0
        self.rate_mod = 0
        self.delay_mod = 0
        self.depth_mod = 0

    def setvalues(self):
        self.waveform = self.read_byte(1) 
        self.rate = self.read_byte(1) 
        self.delay = self.read_byte(1) 
        self.depth = self.read_byte(1) 
        self.rate_mod = self.read_byte(1) 
        self.delay_mod = self.read_byte(1) 
        self.depth_mod = self.read_byte(1) 
 
class LFO1(LFO):
    """
    >>> lfo = LFO1(lfo_sync=1)
    >>> len(lfo.create_chunk())
    22
    """
    def setdefaults(self):
        LFO.setdefaults(self)
        self.lfo_sync = 0
        self.phase = 0 # from sysex docs 
        self.lfo_shift   = 0 # from sysex docs 
        self.lfo_mid_clock_div = 0 # from sysex docs
        self.modwheel = 15
        self.aftertouch = 0

    def create_chunk(self):
        return struct.pack('<4sl14b',
            'lfo ',
            14,
            0,
            self.waveform,
            self.rate,
            self.delay,
            self.depth,
            self.lfo_sync,
            0,
            self.modwheel,
            self.aftertouch,
            self.rate_mod,
            self.delay_mod,
            self.depth_mod,
            0,
            0)


class LFO2(LFO):
    """ 
    >>> lfo = LFO2(lfo_sync=1)
    Traceback (most recent call last):
    Exception: Unknown property: lfo_sync
    >>> lfo = LFO2(lfo_retrigger=1)
    >>> len(lfo.create_chunk())
    22
    """

    def setdefaults(self):
        LFO.setdefaults(self)
        self.lfo_retrigger = 0
        
    def create_chunk(self):
        return struct.pack('<4sl14b',
            'lfo ',
            14,
            0,
            self.waveform,
            self.rate,
            self.delay,
            self.depth,
            0,
            self.lfo_retrigger,
            0,
            0,
            self.rate_mod,
            self.delay_mod,
            self.depth_mod,
            0,
            0)


class Keygroup(Base):
    """ 
    Models a keygroup, a group that contains samples in zones

    >>> kg = Keygroup(blow_note=45)
    Traceback (most recent call last):
    Exception: Unknown property: blow_note
    >>> kg = Keygroup(low_note=45)
    >>> len(kg.create_chunk())
    352
    """
    def setdefaults(self):
        self.low_note=21 
        self.high_note=127 
        self.semi=0
        self.fine=0 
        self.override_fx=0
        self.fx_send=0
        self.pitch_mod1=100
        self.pitch_mod2=0
        self.amp_mod=0
        self.zone_xfade=0
        self.mute_group=0
        self.envelopes = (AmpEnvelope(), FilterEnvelope(), AuxEnvelope())
        self.filter = Filter()
        # always four zones
        self.zones= (Zone(), Zone(), Zone(), Zone(),)

    def create_chunk(self):
        chunk = StringIO()
        chunk.write(struct.pack('<4sl4sl16b',
            'kgrp',
            344,
            'kloc',
            16,
            0,
            0,
            0,
            0,
            self.low_note, 
            self.high_note, 
            self.semi,
            self.fine, 
            self.override_fx,
            self.fx_send,
            self.pitch_mod1,
            self.pitch_mod2,
            self.amp_mod,
            self.zone_xfade,
            self.mute_group,
            0))
        chunk.write(''.join([env.create_chunk() for env in self.envelopes]))
        chunk.write(self.filter.create_chunk()) 
        chunk.write(''.join([zone.create_chunk() for zone in self.zones]))
        return chunk.getvalue()


class Zone(Base):
    """
    >>> z = Zone()
    >>> len(z.create_chunk())
    56
    """
    def setdefaults(self):
        # XXX: 2 bytes missing from description
        self.samplename = '' # no sample assigned 
        self.low_vel = 0
        self.high_vel = 127
        self.fine = 0
        self.semi = 0
        self.filter = 0
        self.pan = 0
        self.playback = 4 # as sample
        self.output = 0 # multi
        self.level = 0
        self.kb_track = 1
        self.velo_start = 0

    def create_chunk(self):
        return struct.pack('<4sl2b20s22bh2b',
            'zone',
            48,
            0,
            len(self.samplename),
            self.samplename,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            self.low_vel,
            self.high_vel,
            self.fine,
            self.semi,
            self.filter,
            self.pan,
            self.playback,
            self.output,
            self.level,
            self.kb_track,
            self.velo_start,
            0,
            0)

class Filter(Base):
    """
    >>> f = Filter(mode=Filter.NOTCH_BI)
    """
    LP2P = 0
    LP4P = 1 
    LP4PPLUS = 2
    BP2P = 3
    BP4P = 4
    BP2PPLUS = 5
    HP1 = 6
    HP2 = 7
    HP8PLUS = 8
    LO_HI = 9
    LO_BAND = 10
    BAND_HI = 11
    NOTCH1 = 12
    NOTCH2 = 13
    NOTCH3 = 14
    NOTCH_WIDE = 15
    NOTCH_BI = 16
    PEAK1 = 17
    PEAK1 = 18
    PEAK1 = 19
    PEAK_WIDE = 20
    PEAK_BI = 21 
    PHASER1 = 22 
    PHASER2 = 23 
    PHASER_BI = 24 
    VOWEL = 25 

    HEADROOM_0db = 0
    HEADROOM_6db = 1
    HEADROOM_12db = 2
    HEADROOM_18db = 3
    HEADROOM_24db = 4
    HEADROOM_30db = 5
    
    def setdefaults(self):
        self.mode = 0 # 2 pole LP
        self.cutoff = 100
        self.resonance = 0
        self.kb_track = 0
        self.mod1_input = 0
        self.mod2_input = 0
        self.mod3_input = 0
        self.headroom = 0

    def create_chunk(self):
        return struct.pack('<4sl10b',
            'filt',
            10,
            0,
            self.mode,
            self.cutoff,
            self.resonance,
            self.kb_track,
            self.mod1_input,
            self.mod2_input,
            self.mod3_input,
            self.headroom,
            0)

class Envelope(Base):
    def setdefaults(self):
        self.attack = 0
        self.decay = 50
        self.release = 15
        self.sustain = 100
        self.velo_to_attack = 0
        self.velo_on_to_release = 0
        self.velo_off_to_release = 0
        self.keyscale = 0


class AmpEnvelope(Envelope):
    """
    >>> amp = AmpEnvelope()
    >>> len(amp.create_chunk()) 
    26
    """
    def create_chunk(self):
        return struct.pack('<4sl18b',
            'env ',
            18,
            0,
            self.attack,
            0,
            self.decay,
            self.release,
            0,
            0,
            self.sustain,
            0,
            0,
            self.velo_to_attack,
            0,
            self.keyscale,
            0,
            self.velo_on_to_release,
            self.velo_off_to_release,
            0,
            0)


class FilterEnvelope(Envelope):
    """
    >>> filt_env = FilterEnvelope()
    >>> len(filt_env.create_chunk()) 
    26
    """
    def setdefaults(self):
        Envelope.setdefaults(self)
        self.filter_env_depth = 0
        
    def create_chunk(self):
        return struct.pack('<4sl18b',
            'env ',
            18,
            0,
            self.attack,
            0,
            self.decay,
            self.release,
            0,
            0,
            self.sustain,
            0,
            self.filter_env_depth,
            self.velo_to_attack,
            0,
            self.keyscale,
            0,
            self.velo_on_to_release,
            self.velo_off_to_release,
            0,
            0)

class AuxEnvelope(Base):
    """
    >>> aux_env = AuxEnvelope()
    >>> len(aux_env.create_chunk()) 
    26
    """
    def setdefaults(self):
        self.rate1 = 0
        self.rate2 = 50
        self.rate3 = 50
        self.rate4 = 15 
        self.level1 = 100 
        self.level2 = 100 
        self.level3 = 100 
        self.level4 = 0 
        self.velo_to_rate1 = 0
        self.kb_to_r2_r4 = 0
        self.velo_on_to_r4 = 0
        self.velo_off_to_r4 = 0
        self.velo_to_out_level = 0

    def create_chunk(self):
        return struct.pack('<4sl18b',
            'env ',
            18,
            0,
            self.rate1,
            self.rate2,
            self.rate3,
            self.rate4,
            self.level1,
            self.level2,
            self.level3,
            self.level4,
            0,
            self.velo_to_rate1,
            0,
            self.kb_to_r2_r4,
            0,
            self.velo_on_to_r4,
            self.velo_off_to_r4,
            self.velo_to_out_level,
            0)

if __name__ == "__main__":
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
    amp = pow(2, 15) - 1
    samplerate = 44100
    no_samples = 1 * samplerate
    loops = (Loop(0, 0, no_samples, 0),)
    freq = 441.0/samplerate
    import math
    samples = []
    for s in range(no_samples):
       samples.append(struct.pack('<h', math.sin(freq * s * 2 * math.pi) * amp))
    samples = ''.join(samples)
    sample = Sample('test', 91, samplerate, loops, samples)
    sample.writefile()




import gtk

def match_func(model, iter, data):
   column, key = data # data is a tuple containing column number, key
   value = model.get_value(iter, column)
   return value == key
def search(model, iter, func, data):
    if not func:
        func = match_func
    while iter:
        if func(model, iter, data):
           return iter
        result = search(model, model.iter_children(iter), func, data)
        if result: return result
        iter = model.iter_next(iter)
    return None

def get_model_from_list(items):
    model = gtk.ListStore(str, str)
    
    if type(items) is dict:
        for i in items.keys():
            s = items[i]
            model.append([i,s])
        return model

    if type(items) is str:
        items = [items,] 
    
    for s in items:
        if type(s) is list:
            model.append(s)
        else:
            model.append([s, s])

    return model

def set_model (cb, model, index=0):
    cb.set_model(model)

    if type(cb) == gtk.ComboBoxEntry:
        cb.set_text_column(0)
    elif type(cb) == gtk.ComboBox:
        cell = gtk.CellRendererText()
        cb.pack_start(cell, True)
        cb.add_attribute(cell, 'text', 0)

def magicCombo(model,select,callback,wrap_width=0):
    cb = gtk.ComboBox()
    cb.set_wrap_width(wrap_width)
    set_model(cb, model, select)
    cb.set_active(select)
    cb.connect('changed', callback)
    return cb




#import os,os.path,re,logging,sys,struct,math,traceback
import gtk,pygtk,gobject
import aksy

import aksui.utils

class SamplerObject(object):
    def __init__(self, s, parent, whichtools, index = None):
        self.parent = parent
        self.s = s
        self.index = index
        self.whichtools = whichtools
        self.specialattrs = []
        self.attrs = []
        self.attrs_minimal = []
        self.attrscache = { }
        self.abbr = {}
       
        # used for alt operations + zones
        self.keygroup_index = None

        # set callback function
        self.set_callback = None

        # set to true if an index needs to be passed before set 
        self.need_index_for_set = False
        
        # set to true if an index needs to be passed in batch operations as an argument
        self.need_index_in_arguments = False
        
        # set to true if this obj needs to be the current one before set
        self.set_current_before_get_set = False
        
    def get_knob(self, attr):
        if attr.startswith("MOD_"):
            pin = self.get_pin_by_name(attr, False)
            if pin:
                return str(pin.level)
            else:
                return ""
        elif attr in self.abbr.keys():
            return self.abbr[attr]
        else:
            return attr
        
    def gettools(self):
        return getattr(self.s, self.whichtools)

    def update(self):
        """ going to replace this w/ a lazy loading type thing
        """
        #print "Update called %s %s." % (self.whichtools, self.get_handle())
        #self.attrscache = { }
        
    def get_handle(self):
        raise NotImplementedError()
        
    def precache(self, clear_cache = False, minimal = True):
        if clear_cache:
            self.attrscache = {}
           
        if minimal and len(self.attrs_minimal) > 0: 
            attrs = self.attrs_minimal
        else:
            attrs = self.attrs
        
        tools = self.gettools()
        handle = self.get_handle()
        assert handle != None
        
        cmds = []
        args = []
        collect = []
        for attr in attrs:
            if attr not in self.specialattrs and attr not in self.attrscache:
                cmd_name = 'get_' + attr + '_cmd'
                cmd = getattr(tools, cmd_name, None)
                if cmd:
                    # if arguments are needed
                    if len(cmd.arg_types) > 0:
                        if self.need_index_in_arguments:
                            collect.append(attr)
                            cmds.append(cmd)
                            args.append([self.index])
                    else:
                        collect.append(attr)
                        cmds.append(cmd)
                        args.append([])

        if len(cmds) > 0:
            if self.keygroup_index != None:
                index = self.keygroup_index
            else:
                index = self.index
                
            results = self.s.execute_alt_request(handle, cmds, args, index)
                
            for i,result in enumerate(results):
                # this assumes the results are in the requested order
                self.attrscache[collect[i]] = result
        
    def set(self, attrname, attrval):
        if self.keygroup_index:
            index = self.keygroup_index
        else:
            index = self.index
            
        handle = self.get_handle()
        assert handle != None

        cmds = []
        args = []
        tools = self.gettools()

        setter_cmd_name = "set_" + attrname + "_cmd"
        set_cmd = getattr(self, setter_cmd_name, getattr(tools, setter_cmd_name))

        cmds.append(set_cmd)
        
        if attrname in self.specialattrs:
            attrval = self.get_special_attr(attrname, attrval)
            
        if self.index != None and self.need_index_for_set:
            arg = [self.index, attrval]
            #print set_cmd.name, repr(set_cmd.id), self.keygroup_index, arg, attrval
            args.append(arg)
        else:
            args.append([attrval])
        
        if len(cmds) > 0:
            self.s.execute_alt_request(handle, cmds, args, self.keygroup_index)
            
            if self.set_callback:
                self.set_callback(attrname, attrval)
            
            # update cache
            self.attrscache[attrname] = attrval

    def get_special_attr(self, attrname, attrval):
        #if attrname == "sample":
        #    attrval = self.samples[attrval]
        return None

    def get(self, attrname):
        pass

    def __getattribute__(self, attrname):
        if attrname.startswith("MOD_") or attrname in object.__getattribute__(self, "attrs") or attrname in object.__getattribute__(self, "specialattrs"):
            cache = object.__getattribute__(self, "attrscache")

            if not attrname in cache and not attrname.startswith("MOD_"):
                tools = self.gettools()

                fname = "get_" + attrname
                func = getattr(tools,fname,None)

                if func:
                    if object.__getattribute__(self,"need_index_for_set"):
                        index = object.__getattribute__(self,"index")
                        if type(index) != int:
                            raise Exception("Index must be an integer, got:" + str(index))
                        else:
                            cache[attrname] = func(index)
                    elif attrname in ["filter","filter_cutoff", "filter_resonance"]:
                        cache[attrname] = func(0) # will expand later for triple filter
                    else:
                        cache[attrname] = func()
                            
                        # TODO: Fix this, level needs a workaround for some reason..
                        if type(cache[attrname]) == tuple:
                            cache[attrname] = cache[attrname][0]
            elif attrname.startswith("MOD_"):
                func = object.__getattribute__(self, "get_pin_by_name")
                pin = func(attrname)
                if pin:
                    cache[attrname] = pin.level
                else:
                    cache[attrname] = None
                    
            if attrname in cache: 
                return cache[attrname]
        #else:
        #    print attrname, object.__getattribute__(self, "attrs")

        return object.__getattribute__(self, attrname)



from aksy.devices.akai.z48.sampler import Z48
from aksy.devices.akai.s56k.sampler import S56K
from aksy.devices.akai.mock_z48.sampler import MockZ48, MockConnector
from aksy.devices.akai.connector import USBConnector
from aksyosc.connector import OSCConnector

_devices = { 'z48': Z48, 's56k': S56K, 'mpc4k': Z48, 'mock_z48': MockZ48 }
_connectors = { 'usb': USBConnector, 'osc': OSCConnector, 'mock': MockConnector }

class Devices:
    """
    return an instance for a specified device
    """

    @staticmethod
    def get_instance(device_id, connector_type='usb', *args, **kwargs):
        try:
            if connector_type == 'usb':
                connector = USBConnector(device_id)
            elif connector_type == 'osc':
                connector = OSCConnector('localhost', 6575)
            elif connector_type == 'mock':
                connector = MockConnector()
            else:
                raise Exception("Unknown connector type ", connector_type)
            return _devices[device_id](connector, *args, **kwargs)
        except KeyError, e:
            raise Exception("Device %s not found" % e[0])




import gtk

class PanelBase(gtk.VBox):
    def __init__(self, so, title = None, set_callback = None):
        gtk.VBox.__init__(self, False, 0)
        if title:
            self.title = title
        else:
            self.title = None
            
        self.set_callback = set_callback
        so.set_callback = set_callback
        self.setup(so)
        
    def clear_children(self,add_label = False):
        for i in (self.get_children()):
            self.remove(i)
            
        if add_label and self.title:
            hbox = gtk.HBox(False,0)
            lb = gtk.Label("<b>%s</b>" % self.title)
            lb.set_justify(gtk.JUSTIFY_LEFT)
            lb.set_use_markup(True)
            hbox.pack_start(lb,False,False,5)
            self.pack_start(hbox,False,False,5)        

    def setup(self,so):
        # sampler
        self.s = so.s
        # samplerobject
        self.so = so



""" Python equivalent of akai section multitools

Methods to manipulate multis
"""

__author__ =  'Walco van Loon'
__version__ =  '0.2'

from aksy.devices.akai.sysex import Command

import aksy.devices.akai.sysex_types

class Multitools:
    def __init__(self, z48):
        self.sampler = z48
        self.get_no_items_cmd = Command('_', '\x18\x01', 'multitools', 'get_no_items', (), None)
        self.get_handles_cmd = Command('_', '\x18\x02\x00', 'multitools', 'get_handles', (), None)
        self.get_names_cmd = Command('_', '\x18\x02\x01', 'multitools', 'get_names', (), None)
        self.get_handles_names_cmd = Command('_', '\x18\x02\x02', 'multitools', 'get_handles_names', (), None)
        self.get_handles_tagged_cmd = Command('_', '\x18\x02\x03', 'multitools', 'get_handles_tagged', (), None)
        self.set_curr_by_handle_cmd = Command('_', '\x18\x03', 'multitools', 'set_curr_by_handle', (aksy.devices.akai.sysex_types.DWORD,), None)
        self.set_curr_by_name_cmd = Command('_', '\x18\x04', 'multitools', 'set_curr_by_name', (aksy.devices.akai.sysex_types.STRING,), None)
        self.get_curr_handle_cmd = Command('_', '\x18\x05', 'multitools', 'get_curr_handle', (), None)
        self.get_curr_name_cmd = Command('_', '\x18\x06', 'multitools', 'get_curr_name', (), None)
        self.get_name_by_handle_cmd = Command('_', '\x18\x07', 'multitools', 'get_name_by_handle', (aksy.devices.akai.sysex_types.DWORD,), None)
        self.get_handle_by_name_cmd = Command('_', '\x18\x08', 'multitools', 'get_handle_by_name', (aksy.devices.akai.sysex_types.STRING,), None)
        self.delete_all_cmd = Command('_', '\x18\x09', 'multitools', 'delete_all', (), None)
        self.delete_curr_cmd = Command('_', '\x18\x0A', 'multitools', 'delete_curr', (), None)
        self.delete_by_handle_cmd = Command('_', '\x18\x0B', 'multitools', 'delete_by_handle', (aksy.devices.akai.sysex_types.DWORD,), None)
        self.rename_curr_cmd = Command('_', '\x18\x0C', 'multitools', 'rename_curr', (aksy.devices.akai.sysex_types.STRING,), None)
        self.rename_by_handle_cmd = Command('_', '\x18\x0D', 'multitools', 'rename_by_handle', (aksy.devices.akai.sysex_types.DWORD, aksy.devices.akai.sysex_types.STRING), None)
        self.tag_cmd = Command('_', '\x18\x0E', 'multitools', 'tag', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.get_tag_bitmap_cmd = Command('_', '\x18\x0F', 'multitools', 'get_tag_bitmap', (), None)
        self.get_curr_modified_cmd = Command('_', '\x18\x10', 'multitools', 'get_curr_modified', (), None)
        self.get_modified_cmd = Command('_', '\x18\x11', 'multitools', 'get_modified', (), None)
        self.delete_tagged_cmd = Command('_', '\x18\x18', 'multitools', 'delete_tagged', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.create_new_cmd = Command('_', '\x18\x40', 'multitools', 'create_new', (aksy.devices.akai.sysex_types.WORD, aksy.devices.akai.sysex_types.STRING), None)
        self.copy_cmd = Command('_', '\x18\x41', 'multitools', 'copy', (aksy.devices.akai.sysex_types.STRING,), None)
        self.delete_part_cmd = Command('_', '\x18\x42', 'multitools', 'delete_part', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.delete_unused_parts_cmd = Command('_', '\x18\x43', 'multitools', 'delete_unused_parts', (), None)
        self.arrange_parts_cmd = Command('_', '\x18\x44', 'multitools', 'arrange_parts', (), None)
        self.set_group_id_cmd = Command('_', '\x1A\x01', 'multitools', 'set_group_id', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_multi_select_method_cmd = Command('_', '\x1A\x02', 'multitools', 'set_multi_select_method', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_multi_select_channel_cmd = Command('_', '\x1A\x03', 'multitools', 'set_multi_select_channel', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_multi_tempo_cmd = Command('_', '\x1A\x04', 'multitools', 'set_multi_tempo', (aksy.devices.akai.sysex_types.WORD,), None)
        self.set_multi_program_no_cmd = Command('_', '\x1A\x08', 'multitools', 'set_multi_program_no', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_multi_part_by_handle_cmd = Command('_', '\x1A\x09', 'multitools', 'set_multi_part_by_handle', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.DWORD), None)
        self.set_multi_part_name_cmd = Command('_', '\x1A\x0A', 'multitools', 'set_multi_part_name', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.STRING), None)
        self.set_no_parts_cmd = Command('_', '\x1A\x0F', 'multitools', 'set_no_parts', (aksy.devices.akai.sysex_types.WORD,), None)
        self.set_part_midi_channel_cmd = Command('_', '\x1A\x10', 'multitools', 'set_part_midi_channel', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.set_part_mute_cmd = Command('_', '\x1A\x11', 'multitools', 'set_part_mute', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BOOL), None)
        self.set_part_solo_cmd = Command('_', '\x1A\x12', 'multitools', 'set_part_solo', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BOOL), None)
        self.set_part_level_cmd = Command('_', '\x1A\x13', 'multitools', 'set_part_level', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.SWORD), None)
        self.set_part_output_cmd = Command('_', '\x1A\x14', 'multitools', 'set_part_output', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.set_part_pan_cmd = Command('_', '\x1A\x15', 'multitools', 'set_part_pan', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.set_part_fx_channel_cmd = Command('_', '\x1A\x16', 'multitools', 'set_part_fx_channel', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.set_part_fx_send_level_cmd = Command('_', '\x1A\x17', 'multitools', 'set_part_fx_send_level', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.SWORD), None)
        self.set_part_cents_tune_cmd = Command('_', '\x1A\x18', 'multitools', 'set_part_cents_tune', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.SWORD), None)
        self.set_part_low_note_cmd = Command('_', '\x1A\x1A', 'multitools', 'set_part_low_note', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.set_part_high_note_cmd = Command('_', '\x1A\x1B', 'multitools', 'set_part_high_note', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.set_part_priority_cmd = Command('_', '\x1A\x1C', 'multitools', 'set_part_priority', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.set_part_program_no_cmd = Command('_', '\x1A\x1D', 'multitools', 'set_part_program_no', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.set_part_group_id_cmd = Command('_', '\x1A\x1F', 'multitools', 'set_part_group_id', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.set_eq_output_channel_cmd = Command('_', '\x1A\x30', 'multitools', 'set_eq_output_channel', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.enable_eq_cmd = Command('_', '\x1A\x31', 'multitools', 'enable_eq', (aksy.devices.akai.sysex_types.BOOL,), None)
        self.set_eq_low_gain_cmd = Command('_', '\x1A\x33', 'multitools', 'set_eq_low_gain', (aksy.devices.akai.sysex_types.SWORD,), None)
        self.set_eq_low_freq_cmd = Command('_', '\x1A\x32', 'multitools', 'set_eq_low_freq', (aksy.devices.akai.sysex_types.WORD,), None)
        self.set_eq_mid_gain_cmd = Command('_', '\x1A\x34', 'multitools', 'set_eq_mid_gain', (aksy.devices.akai.sysex_types.SWORD,), None)
        self.set_eq_mid_freq_cmd = Command('_', '\x1A\x34', 'multitools', 'set_eq_mid_freq', (aksy.devices.akai.sysex_types.WORD,), None)
        self.set_eq_high_gain_cmd = Command('_', '\x1A\x37', 'multitools', 'set_eq_high_gain', (aksy.devices.akai.sysex_types.SWORD,), None)
        self.set_eq_high_freq_cmd = Command('_', '\x1A\x36', 'multitools', 'set_eq_high_freq', (aksy.devices.akai.sysex_types.WORD,), None)
        self.enable_midi_filter_cmd = Command('_', '\x1A\x40', 'multitools', 'enable_midi_filter', (aksy.devices.akai.sysex_types.BOOL,), None)
        self.set_midi_filter_by_part_cmd = Command('_', '\x1A\x41', 'multitools', 'set_midi_filter_by_part', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.set_midi_filter_by_channel_cmd = Command('_', '\x1A\x42', 'multitools', 'set_midi_filter_by_channel', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.set_fx_assign_type_cmd = Command('_', '\x1A\x50', 'multitools', 'set_fx_assign_type', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.set_target_cmd = Command('_', '\x1A\x51', 'multitools', 'set_target', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.WORD), None)
        self.set_destination_cmd = Command('_', '\x1A\x52', 'multitools', 'set_destination', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.WORD), None)
        self.set_change_type_cmd = Command('_', '\x1A\x53', 'multitools', 'set_change_type', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.set_scale_min_cmd = Command('_', '\x1A\x54', 'multitools', 'set_scale_min', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.SBYTE), None)
        self.set_scale_max_cmd = Command('_', '\x1A\x55', 'multitools', 'set_scale_max', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.SBYTE), None)
        self.set_midi_ctl_output_cmd = Command('_', '\x1A\x56', 'multitools', 'set_midi_ctl_output', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.WORD), None)
        self.set_midi_channel_output_cmd = Command('_', '\x1A\x57', 'multitools', 'set_midi_channel_output', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.SBYTE), None)
        self.get_group_id_cmd = Command('_', '\x1B\x01', 'multitools', 'get_group_id', (), None)
        self.get_multi_select_method_cmd = Command('_', '\x1B\x02', 'multitools', 'get_multi_select_method', (), None)
        self.get_multi_select_channel_cmd = Command('_', '\x1B\x03', 'multitools', 'get_multi_select_channel', (), None)
        self.get_multi_tempo_cmd = Command('_', '\x1B\x04', 'multitools', 'get_multi_tempo', (), None)
        self.get_multi_program_no_cmd = Command('_', '\x1B\x08', 'multitools', 'get_multi_program_no', (), None)
        self.get_multi_part_handle_cmd = Command('_', '\x1B\x09', 'multitools', 'get_multi_part_handle', (aksy.devices.akai.sysex_types.DWORD,), None)
        self.get_multi_part_name_cmd = Command('_', '\x1B\x0A', 'multitools', 'get_multi_part_name', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_no_parts_cmd = Command('_', '\x1B\x0F', 'multitools', 'get_no_parts', (), None)
        self.get_part_midi_channel_cmd = Command('_', '\x1B\x10', 'multitools', 'get_part_midi_channel', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_part_mute_cmd = Command('_', '\x1B\x11', 'multitools', 'get_part_mute', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_part_solo_cmd = Command('_', '\x1B\x12', 'multitools', 'get_part_solo', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_part_level_cmd = Command('_', '\x1B\x13', 'multitools', 'get_part_level', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_part_output_cmd = Command('_', '\x1B\x14', 'multitools', 'get_part_output', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_part_pan_cmd = Command('_', '\x1B\x15', 'multitools', 'get_part_pan', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_part_fx_channel_cmd = Command('_', '\x1B\x16', 'multitools', 'get_part_fx_channel', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_part_fx_send_level_cmd = Command('_', '\x1B\x17', 'multitools', 'get_part_fx_send_level', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_part_tune_cmd = Command('_', '\x1B\x18', 'multitools', 'get_part_tune', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_part_low_note_cmd = Command('_', '\x1B\x1A', 'multitools', 'get_part_low_note', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_part_high_note_cmd = Command('_', '\x1B\x1B', 'multitools', 'get_part_high_note', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_part_priority_cmd = Command('_', '\x1B\x1C', 'multitools', 'get_part_priority', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_part_prog_no_cmd = Command('_', '\x1B\x1D', 'multitools', 'get_part_prog_no', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_part_group_id_cmd = Command('_', '\x1B\x1F', 'multitools', 'get_part_group_id', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_eq_output_channel_cmd = Command('_', '\x1B\x30', 'multitools', 'get_eq_output_channel', (), None)
        self.is_eq_enabled_cmd = Command('_', '\x1B\x31', 'multitools', 'is_eq_enabled', (), None)
        self.get_eq_low_gain_cmd = Command('_', '\x1B\x33', 'multitools', 'get_eq_low_gain', (), None)
        self.get_eq_low_freq_cmd = Command('_', '\x1B\x32', 'multitools', 'get_eq_low_freq', (), None)
        self.get_eq_mid_gain_cmd = Command('_', '\x1B\x35', 'multitools', 'get_eq_mid_gain', (), None)
        self.get_eq_mid_freq_cmd = Command('_', '\x1B\x34', 'multitools', 'get_eq_mid_freq', (), None)
        self.get_eq_high_gain_cmd = Command('_', '\x1B\x37', 'multitools', 'get_eq_high_gain', (), None)
        self.get_eq_high_freq_cmd = Command('_', '\x1B\x36', 'multitools', 'get_eq_high_freq', (), None)
        self.is_midi_filter_enabled_cmd = Command('_', '\x1B\x40', 'multitools', 'is_midi_filter_enabled', (), None)
        self.get_midi_filter_by_part_cmd = Command('_', '\x1B\x41', 'multitools', 'get_midi_filter_by_part', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.get_midi_filter_by_channel_cmd = Command('_', '\x1B\x42', 'multitools', 'get_midi_filter_by_channel', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.get_fx_assign_type_cmd = Command('_', '\x1B\x50', 'multitools', 'get_fx_assign_type', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_target_cmd = Command('_', '\x1B\x51', 'multitools', 'get_target', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_destination_cmd = Command('_', '\x1B\x52', 'multitools', 'get_destination', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_change_type_cmd = Command('_', '\x1B\x53', 'multitools', 'get_change_type', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_scale_min_cmd = Command('_', '\x1B\x54', 'multitools', 'get_scale_min', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_scale_max_cmd = Command('_', '\x1B\x55', 'multitools', 'get_scale_max', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_midi_ctl_output_cmd = Command('_', '\x1B\x56', 'multitools', 'get_midi_ctl_output', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.get_midi_channel_output_cmd = Command('_', '\x1B\x57', 'multitools', 'get_midi_channel_output', (aksy.devices.akai.sysex_types.BYTE,), None)

    def get_no_items(self):
        """Get number of items in memory

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_no_items_cmd, ())

    def get_handles(self):
        """Get handles <Data1>: 0=list of handles

        Returns:
            DWORDARRAY
        """
        return self.sampler.execute(self.get_handles_cmd, ())

    def get_names(self):
        """Get names items: <Data1>; 1=list of names;

        Returns:
            STRINGARRAY
        """
        return self.sampler.execute(self.get_names_cmd, ())

    def get_handles_names(self):
        """Get handles names: <Data1>; 2=list of handle+name

        Returns:
            HANDLENAMEARRAY
        """
        return self.sampler.execute(self.get_handles_names_cmd, ())

    def get_handles_tagged(self):
        """Get handles tagged <Data1> ; 3=list of handle+modified/tagged name

        Returns:
            HANDLENAMEARRAY
        """
        return self.sampler.execute(self.get_handles_tagged_cmd, ())

    def set_curr_by_handle(self, arg0):
        """Select by handle
        """
        return self.sampler.execute(self.set_curr_by_handle_cmd, (arg0, ))

    def set_curr_by_name(self, arg0):
        """Select by name
        """
        return self.sampler.execute(self.set_curr_by_name_cmd, (arg0, ))

    def get_curr_handle(self):
        """Get current handle

        Returns:
            DWORD
        """
        return self.sampler.execute(self.get_curr_handle_cmd, ())

    def get_curr_name(self):
        """Get name of current item

        Returns:
            STRING
        """
        return self.sampler.execute(self.get_curr_name_cmd, ())

    def get_name_by_handle(self, arg0):
        """Get item name from handle

        Returns:
            STRING
        """
        return self.sampler.execute(self.get_name_by_handle_cmd, (arg0, ))

    def get_handle_by_name(self, arg0):
        """Get item handle from name

        Returns:
            DWORD
        """
        return self.sampler.execute(self.get_handle_by_name_cmd, (arg0, ))

    def delete_all(self):
        """Delete ALL items from memory
        """
        return self.sampler.execute(self.delete_all_cmd, ())

    def delete_curr(self):
        """Delete current item from memory
        """
        return self.sampler.execute(self.delete_curr_cmd, ())

    def delete_by_handle(self, arg0):
        """Delete item represented by handle <Data1>
        """
        return self.sampler.execute(self.delete_by_handle_cmd, (arg0, ))

    def rename_curr(self, arg0):
        """Rename current item
        """
        return self.sampler.execute(self.rename_curr_cmd, (arg0, ))

    def rename_by_handle(self, arg0, arg1):
        """Rename item represented by handle <Data1>
        """
        return self.sampler.execute(self.rename_by_handle_cmd, (arg0, arg1, ))

    def tag(self, arg0, arg1, arg2):
        """Set Tag Bit <Data1> = bit to set, <Data2> = (0=OFF, 1=ON) <Data3> = (0=CURRENT, 1=ALL)
        """
        return self.sampler.execute(self.tag_cmd, (arg0, arg1, arg2, ))

    def get_tag_bitmap(self):
        """Get Tag Bitmap

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_tag_bitmap_cmd, ())

    def get_curr_modified(self):
        """Get name of current item with modified/tagged info.

        Returns:
            STRINGARRAY
        """
        return self.sampler.execute(self.get_curr_modified_cmd, ())

    def get_modified(self):
        """Get modified state of current item

        Returns:
            BOOL
        """
        return self.sampler.execute(self.get_modified_cmd, ())

    def delete_tagged(self, arg0):
        """Delete tagged items <Data1> = tag bit
        """
        return self.sampler.execute(self.delete_tagged_cmd, (arg0, ))

    def create_new(self, arg0, arg1):
        """Create a new Multi
        """
        return self.sampler.execute(self.create_new_cmd, (arg0, arg1, ))

    def copy(self, arg0):
        """Copy a Multi
        """
        return self.sampler.execute(self.copy_cmd, (arg0, ))

    def delete_part(self, arg0):
        """Delete a specific part
        """
        return self.sampler.execute(self.delete_part_cmd, (arg0, ))

    def delete_unused_parts(self):
        """Delete unused parts
        """
        return self.sampler.execute(self.delete_unused_parts_cmd, ())

    def arrange_parts(self):
        """Arrange parts (sort by MIDI channel)
        """
        return self.sampler.execute(self.arrange_parts_cmd, ())

    def set_group_id(self, arg0):
        """Set Group ID
        """
        return self.sampler.execute(self.set_group_id_cmd, (arg0, ))

    def set_multi_select_method(self, arg0):
        """Set Multi Select <Data1> = (0=OFF, 1=BANK, 2=PROG CHANGE)
        """
        return self.sampler.execute(self.set_multi_select_method_cmd, (arg0, ))

    def set_multi_select_channel(self, arg0):
        """Set Multi Select Channel <Data1> = (1A=0, 2A=1, ..., 16B=31)
        """
        return self.sampler.execute(self.set_multi_select_channel_cmd, (arg0, ))

    def set_multi_tempo(self, arg0):
        """Set Multi Tempo <Data1> = 10 bpm
        """
        return self.sampler.execute(self.set_multi_tempo_cmd, (arg0, ))

    def set_multi_program_no(self, arg0):
        """Set Multi Program Number Data1: (0=OFF, 1=128)
        """
        return self.sampler.execute(self.set_multi_program_no_cmd, (arg0, ))

    def set_multi_part_by_handle(self, arg0, arg1):
        """Set Multi Part by handle <Data1> = Part Number; <Data2> = Handle of program
        """
        return self.sampler.execute(self.set_multi_part_by_handle_cmd, (arg0, arg1, ))

    def set_multi_part_name(self, arg0, arg1):
        """Set Multi Part by name
        """
        return self.sampler.execute(self.set_multi_part_name_cmd, (arg0, arg1, ))

    def set_no_parts(self, arg0):
        """Set number of parts
        """
        return self.sampler.execute(self.set_no_parts_cmd, (arg0, ))

    def set_part_midi_channel(self, arg0, arg1):
        """Set midi channel of part
        """
        return self.sampler.execute(self.set_part_midi_channel_cmd, (arg0, arg1, ))

    def set_part_mute(self, arg0, arg1):
        """Set part mute
        """
        return self.sampler.execute(self.set_part_mute_cmd, (arg0, arg1, ))

    def set_part_solo(self, arg0, arg1):
        """Set part solo
        """
        return self.sampler.execute(self.set_part_solo_cmd, (arg0, arg1, ))

    def set_part_level(self, arg0, arg1):
        """Set part level
        """
        return self.sampler.execute(self.set_part_level_cmd, (arg0, arg1, ))

    def set_part_output(self, arg0, arg1):
        """Set part output
        """
        return self.sampler.execute(self.set_part_output_cmd, (arg0, arg1, ))

    def set_part_pan(self, arg0, arg1):
        """Set part pan
        """
        return self.sampler.execute(self.set_part_pan_cmd, (arg0, arg1, ))

    def set_part_fx_channel(self, arg0, arg1):
        """Set part fx channel
        """
        return self.sampler.execute(self.set_part_fx_channel_cmd, (arg0, arg1, ))

    def set_part_fx_send_level(self, arg0, arg1):
        """Set part fx send level
        """
        return self.sampler.execute(self.set_part_fx_send_level_cmd, (arg0, arg1, ))

    def set_part_cents_tune(self, arg0, arg1):
        """Set part cents tune
        """
        return self.sampler.execute(self.set_part_cents_tune_cmd, (arg0, arg1, ))

    def set_part_low_note(self, arg0, arg1):
        """Set part low note
        """
        return self.sampler.execute(self.set_part_low_note_cmd, (arg0, arg1, ))

    def set_part_high_note(self, arg0, arg1):
        """Set part high note
        """
        return self.sampler.execute(self.set_part_high_note_cmd, (arg0, arg1, ))

    def set_part_priority(self, arg0, arg1):
        """Set part priority
        """
        return self.sampler.execute(self.set_part_priority_cmd, (arg0, arg1, ))

    def set_part_program_no(self, arg0, arg1):
        """Set part program number
        """
        return self.sampler.execute(self.set_part_program_no_cmd, (arg0, arg1, ))

    def set_part_group_id(self, arg0, arg1):
        """Set part group id
        """
        return self.sampler.execute(self.set_part_group_id_cmd, (arg0, arg1, ))

    def set_eq_output_channel(self, arg0):
        """Set EQ Output Channel
        """
        return self.sampler.execute(self.set_eq_output_channel_cmd, (arg0, ))

    def enable_eq(self, arg0):
        """Enable EQ
        """
        return self.sampler.execute(self.enable_eq_cmd, (arg0, ))

    def set_eq_low_gain(self, arg0):
        """Set EQ low gain
        """
        return self.sampler.execute(self.set_eq_low_gain_cmd, (arg0, ))

    def set_eq_low_freq(self, arg0):
        """Set EQ low freq
        """
        return self.sampler.execute(self.set_eq_low_freq_cmd, (arg0, ))

    def set_eq_mid_gain(self, arg0):
        """Set EQ mid gain
        """
        return self.sampler.execute(self.set_eq_mid_gain_cmd, (arg0, ))

    def set_eq_mid_freq(self, arg0):
        """Set EQ mid freq
        """
        return self.sampler.execute(self.set_eq_mid_freq_cmd, (arg0, ))

    def set_eq_high_gain(self, arg0):
        """Set EQ high gain
        """
        return self.sampler.execute(self.set_eq_high_gain_cmd, (arg0, ))

    def set_eq_high_freq(self, arg0):
        """Set EQ high freq
        """
        return self.sampler.execute(self.set_eq_high_freq_cmd, (arg0, ))

    def enable_midi_filter(self, arg0):
        """Enable midi filter
        """
        return self.sampler.execute(self.enable_midi_filter_cmd, (arg0, ))

    def set_midi_filter_by_part(self, arg0, arg1, arg2):
        """Set midi filter by part
        """
        return self.sampler.execute(self.set_midi_filter_by_part_cmd, (arg0, arg1, arg2, ))

    def set_midi_filter_by_channel(self, arg0, arg1, arg2):
        """Set midi filter by channel
        """
        return self.sampler.execute(self.set_midi_filter_by_channel_cmd, (arg0, arg1, arg2, ))

    def set_fx_assign_type(self, arg0, arg1):
        """Set fx assign type (1: FX)
        """
        return self.sampler.execute(self.set_fx_assign_type_cmd, (arg0, arg1, ))

    def set_target(self, arg0, arg1):
        """Set target (part/channel)
        """
        return self.sampler.execute(self.set_target_cmd, (arg0, arg1, ))

    def set_destination(self, arg0, arg1):
        """Set destination
        """
        return self.sampler.execute(self.set_destination_cmd, (arg0, arg1, ))

    def set_change_type(self, arg0, arg1):
        """Set Change type (0: replace, 1: offset)
        """
        return self.sampler.execute(self.set_change_type_cmd, (arg0, arg1, ))

    def set_scale_min(self, arg0, arg1):
        """Set Scale minimum
        """
        return self.sampler.execute(self.set_scale_min_cmd, (arg0, arg1, ))

    def set_scale_max(self, arg0, arg1):
        """Set Scale maximum
        """
        return self.sampler.execute(self.set_scale_max_cmd, (arg0, arg1, ))

    def set_midi_ctl_output(self, arg0, arg1):
        """Set Midi controller output (0: off, 1-128)
        """
        return self.sampler.execute(self.set_midi_ctl_output_cmd, (arg0, arg1, ))

    def set_midi_channel_output(self, arg0, arg1):
        """Set Midi channel output (0: off, 1-128)
        """
        return self.sampler.execute(self.set_midi_channel_output_cmd, (arg0, arg1, ))

    def get_group_id(self):
        """Get Group ID

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_group_id_cmd, ())

    def get_multi_select_method(self):
        """Get Multi Select <Reply1> = (0=OFF, 1=BANK, 2=PROG CHANGE)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_multi_select_method_cmd, ())

    def get_multi_select_channel(self):
        """Get Multi Select Channel <Reply1> = (1A=0, 2A=1, ..., 16B=31)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_multi_select_channel_cmd, ())

    def get_multi_tempo(self):
        """Get Multi Tempo <Reply> = 10 bpm

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_multi_tempo_cmd, ())

    def get_multi_program_no(self):
        """Get Multi Program Number

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_multi_program_no_cmd, ())

    def get_multi_part_handle(self, arg0):
        """Get Multi Part handle. <Data1> = Part Number;<Reply> = Handle of program>

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_multi_part_handle_cmd, (arg0, ))

    def get_multi_part_name(self, arg0):
        """Get Multi Part name. <Data1> = Part Number; <Reply> = Name of part

        Returns:
            STRING
        """
        return self.sampler.execute(self.get_multi_part_name_cmd, (arg0, ))

    def get_no_parts(self):
        """Get Number of Parts. <Reply> = new number of parts

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_no_parts_cmd, ())

    def get_part_midi_channel(self, arg0):
        """Get Part MIDI Channel, (Data1 = Part Number-1a) <Reply> = (1A=0, 2A=1, ..., 16B=31)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_part_midi_channel_cmd, (arg0, ))

    def get_part_mute(self, arg0):
        """Get Part Mute, <Reply> = (0=OFF, 1=ON)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_part_mute_cmd, (arg0, ))

    def get_part_solo(self, arg0):
        """Get Part Solo, <Reply> = (0=OFF, 1=ON)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_part_solo_cmd, (arg0, ))

    def get_part_level(self, arg0):
        """Get Part Level, <Reply> = PartLevel in 10 dB

        Returns:
            SWORD
        """
        return self.sampler.execute(self.get_part_level_cmd, (arg0, ))

    def get_part_output(self, arg0):
        """Get Part Output, <Reply> = (Output: 0 = L/R; 1-4 = op1/2 op7/8; 5 = 14 = L, R, op1-op8)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_part_output_cmd, (arg0, ))

    def get_part_pan(self, arg0):
        """Get Part Pan/Balance, <Reply> = Pan/Bal (0-100 = L50-R50); centre=50

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_part_pan_cmd, (arg0, ))

    def get_part_fx_channel(self, arg0):
        """Get Part Effects Channel: Reply = (0=OFF, 1=FX1, 2=FX2, 3=RV3, 4=RV4)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_part_fx_channel_cmd, (arg0, ))

    def get_part_fx_send_level(self, arg0):
        """Get Part FX Send Level <Reply> = level in 10 dB

        Returns:
            SWORD
        """
        return self.sampler.execute(self.get_part_fx_send_level_cmd, (arg0, ))

    def get_part_tune(self, arg0):
        """Get Part Cents Tune

        Returns:
            SWORD
        """
        return self.sampler.execute(self.get_part_tune_cmd, (arg0, ))

    def get_part_low_note(self, arg0):
        """Get Part Low Note

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_part_low_note_cmd, (arg0, ))

    def get_part_high_note(self, arg0):
        """Get Part High Note

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_part_high_note_cmd, (arg0, ))

    def get_part_priority(self, arg0):
        """Get Part Priority

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_part_priority_cmd, (arg0, ))

    def get_part_prog_no(self, arg0):
        """Get Part Program Number

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_part_prog_no_cmd, (arg0, ))

    def get_part_group_id(self, arg0):
        """Get Part Group ID

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_part_group_id_cmd, (arg0, ))

    def get_eq_output_channel(self):
        """Get EQ Output Channel

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_eq_output_channel_cmd, ())

    def is_eq_enabled(self):
        """Whether EQ is enabled

        Returns:
            BOOL
        """
        return self.sampler.execute(self.is_eq_enabled_cmd, ())

    def get_eq_low_gain(self):
        """Get EQ low gain

        Returns:
            SWORD
        """
        return self.sampler.execute(self.get_eq_low_gain_cmd, ())

    def get_eq_low_freq(self):
        """Get EQ low freq

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_eq_low_freq_cmd, ())

    def get_eq_mid_gain(self):
        """Get EQ mid gain

        Returns:
            SWORD
        """
        return self.sampler.execute(self.get_eq_mid_gain_cmd, ())

    def get_eq_mid_freq(self):
        """Get EQ mid freq

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_eq_mid_freq_cmd, ())

    def get_eq_high_gain(self):
        """Get EQ high gain

        Returns:
            SWOR
        """
        return self.sampler.execute(self.get_eq_high_gain_cmd, ())

    def get_eq_high_freq(self):
        """Get EQ high freq

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_eq_high_freq_cmd, ())

    def is_midi_filter_enabled(self):
        """Whether the midi filter is enabled

        Returns:
            BOOL
        """
        return self.sampler.execute(self.is_midi_filter_enabled_cmd, ())

    def get_midi_filter_by_part(self, arg0, arg1):
        """Get midi filter by part

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_midi_filter_by_part_cmd, (arg0, arg1, ))

    def get_midi_filter_by_channel(self, arg0, arg1):
        """Get midi filter by channel

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_midi_filter_by_channel_cmd, (arg0, arg1, ))

    def get_fx_assign_type(self, arg0):
        """Get fx assign type (1: FX)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_fx_assign_type_cmd, (arg0, ))

    def get_target(self, arg0):
        """Get target (part/channel)

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_target_cmd, (arg0, ))

    def get_destination(self, arg0):
        """Get destination

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_destination_cmd, (arg0, ))

    def get_change_type(self, arg0):
        """Get Change type (0: replace, 1: offset)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_change_type_cmd, (arg0, ))

    def get_scale_min(self, arg0):
        """Get Scale minimum

        Returns:
            SBYTE
        """
        return self.sampler.execute(self.get_scale_min_cmd, (arg0, ))

    def get_scale_max(self, arg0):
        """Get Scale maximum

        Returns:
            SBYTE
        """
        return self.sampler.execute(self.get_scale_max_cmd, (arg0, ))

    def get_midi_ctl_output(self, arg0):
        """Get Midi controller output (0: off, 1-128)

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_midi_ctl_output_cmd, (arg0, ))

    def get_midi_channel_output(self, arg0):
        """Get Midi channel output (0: off, 1-128)

        Returns:
            SBYTE
        """
        return self.sampler.execute(self.get_midi_channel_output_cmd, (arg0, ))




class ModulationMatrix:
    # should be a total of 36 (0 based) * marks sample + hold mods
    sources      = ['NONE','MODWHEEL','BEND UP','BEND DOWN',
                    'AFTERTOUCH','VELOCITY','BIPOLAR VELOCITY','OFF VELOCITY',
                    'KEYBOARD','LFO1','LFO2','AMP ENV',
                    'FILTER ENV','AUX ENV','MIDI CTRL','Q-LINK 1',
                    'Q-LINK 2','Q-LINK 3','Q-LINK 4','Q-LINK 5',
                    'Q-LINK 6','Q-LINK 7','Q-LINK 8','*MODWHEEL',
                    '*BEND UP','*BEND DOWN','*Q-LINK 1','*Q-LINK 2',
                    '*Q-LINK 3','*Q-LINK 4','*Q-LINK 5','*Q-LINK 6',
                    '*Q-LINK 7','*Q-LINK 8','*LFO1','*LFO2','*MIDI CTRL']

    # should be a total of 52 (0 based), qlink adds 4 'EXT' options
    destinations = ['NONE',           'AMPLITUDE',    'PAN',            'PITCH',
                    'LFO1 RATE',      'LFO1 DEPTH',   'LFO1 DELAY',     'LFO1 PHASE',
                    'LFO1 OFFSET',    'LFO2 RATE',    'LFO2 DEPTH',     'LFO2 DELAY',
                    'LFO2 PHASE',     'LFO2 OFFSET',  'CUTOFF',  'RESONANCE',
                    'TR 1 CUTOFF',    'TR 1 RES',     'TR 2 CUTOFF',    'TR 2 RES',
                    'TR 3 CUTOFF',    'TR 3 RES',     'AMP ENV ATTACK', 'AMP ENV DECAY',
                    'AMP ENV RELEASE','FILT ENV R1',  'FILT ENV R2',    'FILT ENV R4',
                    'AUX ENV R1',     'AUX ENV R2',   'AUX ENV R4',     'ZONE CROSSFADE',
                    'ZONE 1 LEVEL',   'ZONE 1 PAN',   'ZONE 1 PITCH',   'ZONE 1 START', 'ZONE 1 FILTER'
                    'ZONE 2 LEVEL',   'ZONE 2 PAN',   'ZONE 2 PITCH',   'ZONE 2 START', 'ZONE 2 FILTER'
                    'ZONE 3 LEVEL',   'ZONE 3 PAN',   'ZONE 3 PITCH',   'ZONE 3 START', 'ZONE 3 FILTER'
                    'ZONE 4 LEVEL',   'ZONE 4 PAN',   'ZONE 4 PITCH',   'ZONE 4 START', 'ZONE 4 FILTER']

    def __init__(self,program,index):
        pass




import gtk

from aksui.UI import base, editors

class MultiEditorWindow(gtk.Window):
    def __init__(self, s, m):
        gtk.Window.__init__(self)
        self.s = s
        self.setup(m)
        
    def setup(self, m):
        self.set_default_size(600, 500)
        self.editor = MultiEditor(self.s, m)
        self.add(self.editor.editor)
        self.set_title("aksui: %s" % (m.name))
        #self.editor.setup(p)
        
class MultiEditor(base.Base):
    def __init__(self, s, m):
        base.Base.__init__(self, m, "multiEditorZ")
        self.s = m.s
        self.multiEditorVBox = editors.MultiEditorVBox(s, m)
        self.w_viewportMultis.add(self.multiEditorVBox)



import samplerobject

class Zone(samplerobject.SamplerObject):
    def __init__(self, kg, index):
        samplerobject.SamplerObject.__init__(self, kg.s, kg, "zonetools", index)
        self.samples = kg.samples
        self.keygroup = kg
        self.specialattrs = ["sample",]
        self.keygroup_index = self.keygroup.index
        self.need_index_for_set = True
        self.need_index_in_arguments = True
        self.set_current_before_get_set = True

        # skip 'sample' for now
        self.attrs = ["sample", "level", "pan", "output", "filter", "tune", "keyboard_track", "playback", "mod_start", "low_velocity", "high_velocity", "mute"]
        self.abbr = {'mod_start' : 'start', 'low_velocity' : 'lo', 'high_velocity' : 'hi'}

        #self.precache()
        
    def get_handle(self):
        return self.keygroup.p.handle
        
    def get_special_attr(self,attrname,attrval):
        # actually used for setting
        if attrname == "sample":
            return attrval
        else:
            return None
